from __future__ import annotations

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, TypeVar

from infrahub.message_bus.messages import ROUTING_KEY_MAP

ResponseClass = TypeVar("ResponseClass")

if TYPE_CHECKING:
    from infrahub.message_bus import InfrahubMessage, InfrahubResponse
    from infrahub.message_bus.types import MessageTTL
    from infrahub.services import InfrahubServices


class InfrahubMessageBus(ABC):
    DELIVER_TIMEOUT: int = 30 * 60  # 30 minutes
    worker_bindings: list[str] = [
        "check.*.*",
        "event.*.*",
        "finalize.*.*",
        "git.*.*",
        "refresh.webhook.*",
        "request.*.*",
        "send.*.*",
        "schema.*.*",
        "transform.*.*",
        "trigger.*.*",
    ]
    event_bindings: list[str] = ["refresh.registry.*"]
    broadcasted_event_bindings: list[str] = ["refresh.git.*"]
    service: InfrahubServices

    async def shutdown(self) -> None:  # noqa: B027 We want a default empty behavior, so it's ok to have an empty non-abstract method.
        """Shutdown the Message bus"""

    @abstractmethod
    async def publish(
        self, message: InfrahubMessage, routing_key: str, delay: MessageTTL | None = None, is_retry: bool = False
    ) -> None:
        raise NotImplementedError()

    @abstractmethod
    async def reply(self, message: InfrahubMessage, routing_key: str) -> None:
        raise NotImplementedError()

    @abstractmethod
    async def rpc(self, message: InfrahubMessage, response_class: type[ResponseClass]) -> ResponseClass:
        raise NotImplementedError()

    async def send(self, message: InfrahubMessage, delay: MessageTTL | None = None, is_retry: bool = False) -> None:
        routing_key = ROUTING_KEY_MAP.get(type(message))
        if not routing_key:
            raise ValueError("Unable to determine routing key")
        await self.publish(message, routing_key=routing_key, delay=delay, is_retry=is_retry)

    # TODO rename it
    async def reply_if_initiator_meta(self, message: InfrahubResponse, initiator: InfrahubMessage) -> None:
        if initiator.meta:
            message.meta.correlation_id = initiator.meta.correlation_id
            routing_key = initiator.meta.reply_to or ""
            await self.reply(message, routing_key=routing_key)



from infrahub.core.constants import (
    BranchSupportType,
)

builtin_tag = {
    "name": "Tag",
    "namespace": "Builtin",
    "description": "Standard Tag object to attached to other objects to provide some context.",
    "include_in_menu": True,
    "icon": "mdi:tag-multiple",
    "label": "Tag",
    "default_filter": "name__value",
    "order_by": ["name__value"],
    "display_labels": ["name__value"],
    "branch": BranchSupportType.AWARE.value,
    "uniqueness_constraints": [["name__value"]],
    "attributes": [
        {"name": "name", "kind": "Text", "unique": True},
        {"name": "description", "kind": "Text", "optional": True},
    ],
}



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from infrahub.message_bus.types import KVTTL


class InfrahubCache(ABC):
    """Base class for caching services"""

    @abstractmethod
    async def delete(self, key: str) -> None:
        """Delete a key from the cache."""
        raise NotImplementedError()

    @abstractmethod
    async def get(self, key: str) -> str | None:
        """Retrieve a value from the cache."""
        raise NotImplementedError()

    @abstractmethod
    async def get_values(self, keys: list[str]) -> list[str | None]:
        """Return a list the values for requested keys."""
        raise NotImplementedError()

    @abstractmethod
    async def list_keys(self, filter_pattern: str) -> list[str]:
        """Return a list of active keys that match the provided filter."""
        raise NotImplementedError()

    @abstractmethod
    async def set(self, key: str, value: str, expires: KVTTL | None = None, not_exists: bool = False) -> bool | None:
        """Set a value in the cache."""
        raise NotImplementedError()



from infrahub.core.constants import (
    BranchSupportType,
    InfrahubKind,
)

core_check_definition = {
    "name": "CheckDefinition",
    "namespace": "Core",
    "include_in_menu": False,
    "icon": "mdi:check-all",
    "label": "Check Definition",
    "default_filter": "name__value",
    "order_by": ["name__value"],
    "display_labels": ["name__value"],
    "branch": BranchSupportType.AWARE.value,
    "uniqueness_constraints": [["name__value"]],
    "generate_profile": False,
    "inherit_from": [InfrahubKind.TASKTARGET],
    "attributes": [
        {"name": "name", "kind": "Text", "unique": True},
        {"name": "description", "kind": "Text", "optional": True},
        {"name": "file_path", "kind": "Text"},
        {"name": "class_name", "kind": "Text"},
        {"name": "timeout", "kind": "Number", "default_value": 10},
        {"name": "parameters", "kind": "JSON", "optional": True},
    ],
    "relationships": [
        {
            "name": "repository",
            "peer": InfrahubKind.GENERICREPOSITORY,
            "kind": "Attribute",
            "cardinality": "one",
            "identifier": "check_definition__repository",
            "optional": False,
        },
        {
            "name": "query",
            "peer": InfrahubKind.GRAPHQLQUERY,
            "kind": "Attribute",
            "identifier": "check_definition__graphql_query",
            "cardinality": "one",
            "optional": True,
        },
        {
            "name": "targets",
            "peer": InfrahubKind.GENERICGROUP,
            "kind": "Attribute",
            "identifier": "check_definition___group",
            "cardinality": "one",
            "optional": True,
        },
        {
            "name": "tags",
            "peer": InfrahubKind.TAG,
            "kind": "Attribute",
            "optional": True,
            "cardinality": "many",
        },
    ],
}



from infrahub.core.constants import (
    BranchSupportType,
    InfrahubKind,
    RelationshipDeleteBehavior,
)
from infrahub.proposed_change.constants import ProposedChangeState

core_proposed_change = {
    "name": "ProposedChange",
    "namespace": "Core",
    "description": "Metadata related to a proposed change",
    "include_in_menu": False,
    "icon": "mdi:file-replace-outline",
    "label": "Proposed Change",
    "default_filter": "name__value",
    "display_labels": ["name__value"],
    "generate_profile": False,
    "branch": BranchSupportType.AGNOSTIC.value,
    "inherit_from": [InfrahubKind.TASKTARGET],
    "documentation": "/topics/proposed-change",
    "attributes": [
        {"name": "name", "kind": "Text", "optional": False},
        {"name": "description", "kind": "TextArea", "optional": True},
        {"name": "source_branch", "kind": "Text", "optional": False},
        {"name": "destination_branch", "kind": "Text", "optional": False},
        {
            "name": "state",
            "kind": "Text",
            "enum": ProposedChangeState.available_types(),
            "default_value": ProposedChangeState.OPEN.value,
            "optional": True,
        },
    ],
    "relationships": [
        {
            "name": "approved_by",
            "peer": InfrahubKind.GENERICACCOUNT,
            "optional": True,
            "cardinality": "many",
            "kind": "Attribute",
            "branch": BranchSupportType.AGNOSTIC.value,
            "identifier": "coreaccount__proposedchange_approved_by",
        },
        {
            "name": "reviewers",
            "peer": InfrahubKind.GENERICACCOUNT,
            "optional": True,
            "kind": "Attribute",
            "cardinality": "many",
            "branch": BranchSupportType.AGNOSTIC.value,
            "identifier": "coreaccount__proposedchange_reviewed_by",
        },
        {
            "name": "created_by",
            "peer": InfrahubKind.GENERICACCOUNT,
            "optional": True,
            "cardinality": "one",
            "branch": BranchSupportType.AGNOSTIC.value,
            "identifier": "coreaccount__proposedchange_created_by",
        },
        {
            "name": "comments",
            "peer": InfrahubKind.CHANGECOMMENT,
            "kind": "Component",
            "optional": True,
            "cardinality": "many",
            "on_delete": RelationshipDeleteBehavior.CASCADE,
        },
        {
            "name": "threads",
            "peer": InfrahubKind.THREAD,
            "identifier": "proposedchange__thread",
            "kind": "Component",
            "optional": True,
            "cardinality": "many",
            "on_delete": RelationshipDeleteBehavior.CASCADE,
        },
        {
            "name": "validations",
            "peer": InfrahubKind.VALIDATOR,
            "kind": "Component",
            "identifier": "proposed_change__validator",
            "optional": True,
            "cardinality": "many",
            "on_delete": RelationshipDeleteBehavior.CASCADE,
        },
    ],
}



from __future__ import annotations

from typing import TYPE_CHECKING, Any

from graphene import Boolean, InputObjectType, Mutation, String

from infrahub.core.account import ObjectPermission
from infrahub.core.constants import ComputedAttributeKind, MutationAction, PermissionAction, PermissionDecision
from infrahub.core.manager import NodeManager
from infrahub.core.registry import registry
from infrahub.database import retry_db_transaction
from infrahub.events import EventMeta
from infrahub.events.node_action import NodeMutatedEvent
from infrahub.exceptions import NodeNotFoundError, ValidationError
from infrahub.graphql.context import apply_external_context
from infrahub.graphql.types.context import ContextInput
from infrahub.log import get_log_data
from infrahub.worker import WORKER_IDENTITY

if TYPE_CHECKING:
    from graphql import GraphQLResolveInfo

    from infrahub.graphql.initialization import GraphqlContext


class InfrahubComputedAttributeUpdateInput(InputObjectType):
    id = String(required=True)
    kind = String(required=True)
    attribute = String(required=True)
    value = String(required=True)


class UpdateComputedAttribute(Mutation):
    class Arguments:
        data = InfrahubComputedAttributeUpdateInput(required=True)
        context = ContextInput(required=False)

    ok = Boolean()

    @classmethod
    @retry_db_transaction(name="update_computed_attribute")
    async def mutate(
        cls,
        _: dict,
        info: GraphQLResolveInfo,
        data: InfrahubComputedAttributeUpdateInput,
        context: ContextInput | None = None,
    ) -> UpdateComputedAttribute:
        graphql_context: GraphqlContext = info.context
        node_schema = registry.schema.get_node_schema(
            name=str(data.kind), branch=graphql_context.branch.name, duplicate=False
        )
        target_attribute = node_schema.get_attribute(name=str(data.attribute))
        if (
            not target_attribute.computed_attribute
            or target_attribute.computed_attribute.kind == ComputedAttributeKind.USER
        ):
            raise ValidationError(input_value=f"{node_schema.kind}.{target_attribute.name} is not a computed attribute")

        graphql_context.active_permissions.raise_for_permission(
            permission=ObjectPermission(
                namespace=node_schema.namespace,
                name=node_schema.name,
                action=PermissionAction.UPDATE.value,
                decision=PermissionDecision.ALLOW_DEFAULT.value
                if graphql_context.branch.name == registry.default_branch
                else PermissionDecision.ALLOW_OTHER.value,
            )
        )
        await apply_external_context(graphql_context=graphql_context, context_input=context)

        if not (
            target_node := await NodeManager.get_one(
                db=graphql_context.db, kind=node_schema.kind, id=str(data.id), branch=graphql_context.branch
            )
        ):
            raise NodeNotFoundError(
                node_type="target_node",
                identifier=str(data.id),
                message="The indicated node was not found in the database",
            )

        attribute_field = getattr(target_node, str(data.attribute), None)
        if not attribute_field:
            raise NodeNotFoundError(
                node_type="target_node",
                identifier=str(data.id),
                message="The indicated not does not have the specified attribute_name",
            )
        if attribute_field.value != str(data.value):
            attribute_field.value = str(data.value)
            async with graphql_context.db.start_transaction() as dbt:
                await target_node.save(db=dbt, fields=[str(data.attribute)])

            log_data = get_log_data()
            request_id = log_data.get("request_id", "")

            event = NodeMutatedEvent(
                kind=node_schema.kind,
                node_id=target_node.get_id(),
                data=target_node.node_changelog.model_dump(),
                fields=[str(data.attribute)],
                action=MutationAction.UPDATED,
                meta=EventMeta(
                    context=graphql_context.get_context(),
                    initiator_id=WORKER_IDENTITY,
                    request_id=request_id,
                    account_id=graphql_context.active_account_session.account_id,
                    branch=graphql_context.branch,
                ),
            )
            await graphql_context.active_service.event.send(event=event)

        result: dict[str, Any] = {"ok": True}

        return cls(**result)



from __future__ import annotations

from typing import Any
from uuid import uuid4

import pydantic
from prefect import flow, get_run_logger
from prefect.client.schemas.objects import State  # noqa: TC002
from prefect.states import Completed, Failed

from infrahub import lock
from infrahub.context import InfrahubContext  # noqa: TC001  needed for prefect flow
from infrahub.core import registry
from infrahub.core.branch import Branch
from infrahub.core.changelog.diff import DiffChangelogCollector, MigrationTracker
from infrahub.core.constants import MutationAction
from infrahub.core.diff.coordinator import DiffCoordinator
from infrahub.core.diff.ipam_diff_parser import IpamDiffParser
from infrahub.core.diff.merger.merger import DiffMerger
from infrahub.core.diff.model.path import BranchTrackingId, EnrichedDiffRoot, EnrichedDiffRootMetadata
from infrahub.core.diff.repository.repository import DiffRepository
from infrahub.core.merge import BranchMerger
from infrahub.core.migrations.schema.models import SchemaApplyMigrationData
from infrahub.core.migrations.schema.tasks import schema_apply_migrations
from infrahub.core.timestamp import Timestamp
from infrahub.core.validators.determiner import ConstraintValidatorDeterminer
from infrahub.core.validators.models.validate_migration import SchemaValidateMigrationData
from infrahub.core.validators.tasks import schema_validate_migrations
from infrahub.dependencies.registry import get_component_registry
from infrahub.events.branch_action import BranchCreatedEvent, BranchDeletedEvent, BranchMergedEvent, BranchRebasedEvent
from infrahub.events.models import EventMeta, InfrahubEvent
from infrahub.events.node_action import NodeMutatedEvent
from infrahub.exceptions import BranchNotFoundError, MergeFailedError, ValidationError
from infrahub.graphql.mutations.models import BranchCreateModel  # noqa: TC001
from infrahub.log import get_log_data
from infrahub.message_bus import Meta, messages
from infrahub.services import InfrahubServices  # noqa: TC001  needed for prefect flow
from infrahub.worker import WORKER_IDENTITY
from infrahub.workflows.catalogue import (
    BRANCH_CANCEL_PROPOSED_CHANGES,
    DIFF_REFRESH_ALL,
    GIT_REPOSITORIES_CREATE_BRANCH,
    IPAM_RECONCILIATION,
)
from infrahub.workflows.utils import add_tags


@flow(name="branch-rebase", flow_run_name="Rebase branch {branch}")
async def rebase_branch(branch: str, context: InfrahubContext, service: InfrahubServices) -> None:
    async with service.database.start_session() as db:
        log = get_run_logger()
        await add_tags(branches=[branch])
        obj = await Branch.get_by_name(db=db, name=branch)
        base_branch = await Branch.get_by_name(db=db, name=registry.default_branch)
        component_registry = get_component_registry()
        diff_repository = await component_registry.get_component(DiffRepository, db=db, branch=obj)
        diff_coordinator = await component_registry.get_component(DiffCoordinator, db=db, branch=obj)
        diff_merger = await component_registry.get_component(DiffMerger, db=db, branch=obj)
        initial_from_time = Timestamp(obj.get_branched_from())
        merger = BranchMerger(
            db=db,
            diff_coordinator=diff_coordinator,
            diff_merger=diff_merger,
            diff_repository=diff_repository,
            source_branch=obj,
            service=service,
        )

        enriched_diff_metadata = await diff_coordinator.update_branch_diff(base_branch=base_branch, diff_branch=obj)
        async for _ in diff_repository.get_all_conflicts_for_diff(
            diff_branch_name=enriched_diff_metadata.diff_branch_name, diff_id=enriched_diff_metadata.uuid
        ):
            # if there are any conflicts, raise the error
            raise ValidationError(
                f"Branch {obj.name} contains conflicts with the default branch that must be addressed."
                " Please review the diff for details and manually update the conflicts before rebasing."
            )
        node_diff_field_summaries = await diff_repository.get_node_field_summaries(
            diff_branch_name=enriched_diff_metadata.diff_branch_name, diff_id=enriched_diff_metadata.uuid
        )

        candidate_schema = merger.get_candidate_schema()
        determiner = ConstraintValidatorDeterminer(schema_branch=candidate_schema)
        constraints = await determiner.get_constraints(node_diffs=node_diff_field_summaries)

        # If there are some changes related to the schema between this branch and main, we need to
        #  - Run all the validations to ensure everything is correct before rebasing the branch
        #  - Run all the migrations after the rebase
        if obj.has_schema_changes:
            constraints += await merger.calculate_validations(target_schema=candidate_schema)
        if constraints:
            responses = await schema_validate_migrations(
                message=SchemaValidateMigrationData(
                    branch=obj, schema_branch=candidate_schema, constraints=constraints
                ),
                service=service,
            )
            error_messages = [violation.message for response in responses for violation in response.violations]
            if error_messages:
                raise ValidationError(",\n".join(error_messages))

        schema_in_main_before = merger.destination_schema.duplicate()
        migrations = []
        async with lock.registry.global_graph_lock():
            async with db.start_transaction() as dbt:
                await obj.rebase(db=dbt)
                log.info("Branch successfully rebased")

            if obj.has_schema_changes:
                # NOTE there is a bit additional work in order to calculate a proper diff that will
                # allow us to pull only the part of the schema that has changed, for now the safest option is to pull
                # Everything
                # schema_diff = await merger.has_schema_changes()
                # TODO Would be good to convert this part to a Prefect Task in order to track it properly
                updated_schema = await registry.schema.load_schema_from_db(
                    db=db,
                    branch=obj,
                    # schema=merger.source_schema.duplicate(),
                    # schema_diff=schema_diff,
                )
                registry.schema.set_schema_branch(name=obj.name, schema=updated_schema)
                obj.update_schema_hash()
                await obj.save(db=db)

                # Execute the migrations
                migrations = await merger.calculate_migrations(target_schema=updated_schema)

                errors = await schema_apply_migrations(
                    message=SchemaApplyMigrationData(
                        branch=merger.source_branch,
                        new_schema=candidate_schema,
                        previous_schema=schema_in_main_before,
                        migrations=migrations,
                    ),
                    service=service,
                )
                for error in errors:
                    log.error(error)

        default_branch_diff = await _get_diff_root(
            diff_coordinator=diff_coordinator,
            enriched_diff_metadata=enriched_diff_metadata,
            diff_repository=diff_repository,
            base_branch=base_branch,
            target_from=initial_from_time,
        )

        # -------------------------------------------------------------
        # Trigger the reconciliation of IPAM data after the rebase
        # -------------------------------------------------------------
        diff_parser = await component_registry.get_component(IpamDiffParser, db=db, branch=obj)
        ipam_node_details = await diff_parser.get_changed_ipam_node_details(
            source_branch_name=obj.name,
            target_branch_name=registry.default_branch,
        )
        if ipam_node_details:
            await service.workflow.submit_workflow(
                workflow=IPAM_RECONCILIATION,
                context=context,
                parameters={"branch": obj.name, "ipam_node_details": ipam_node_details},
            )

    await service.workflow.submit_workflow(
        workflow=DIFF_REFRESH_ALL, context=context, parameters={"branch_name": obj.name}
    )

    # -------------------------------------------------------------
    # Generate an event to indicate that a branch has been rebased
    # -------------------------------------------------------------
    rebase_event = BranchRebasedEvent(
        branch_name=obj.name, branch_id=str(obj.uuid), meta=EventMeta(branch=obj, context=context)
    )
    events: list[InfrahubEvent] = [rebase_event]
    changelog_collector = DiffChangelogCollector(
        diff=default_branch_diff, branch=obj, db=db, migration_tracker=MigrationTracker(migrations=migrations)
    )
    for action, node_changelog in changelog_collector.collect_changelogs():
        mutate_event = NodeMutatedEvent(
            kind=node_changelog.node_kind,
            node_id=node_changelog.node_id,
            data=node_changelog,
            action=MutationAction.from_diff_action(diff_action=action),
            fields=node_changelog.updated_fields,
            meta=EventMeta.from_parent(parent=rebase_event, branch=obj),
        )
        events.append(mutate_event)

    for event in events:
        await service.event.send(event)


@flow(name="branch-merge", flow_run_name="Merge branch {branch} into main")
async def merge_branch(branch: str, context: InfrahubContext, service: InfrahubServices) -> None:
    async with service.database.start_session() as db:
        log = get_run_logger()

        await add_tags(branches=[branch, registry.default_branch])

        obj = await Branch.get_by_name(db=db, name=branch)
        default_branch = await registry.get_branch(db=db, branch=registry.default_branch)
        component_registry = get_component_registry()
        merge_event = BranchMergedEvent(
            branch_name=obj.name,
            branch_id=str(obj.get_uuid()),
            meta=EventMeta.from_context(context=context, branch=registry.get_global_branch()),
        )

        merger: BranchMerger | None = None
        async with lock.registry.global_graph_lock():
            # await update_diff(model=RequestDiffUpdate(branch_name=obj.name))

            diff_repository = await component_registry.get_component(DiffRepository, db=db, branch=obj)
            diff_coordinator = await component_registry.get_component(DiffCoordinator, db=db, branch=obj)
            diff_merger = await component_registry.get_component(DiffMerger, db=db, branch=obj)
            merger = BranchMerger(
                db=db,
                diff_coordinator=diff_coordinator,
                diff_merger=diff_merger,
                diff_repository=diff_repository,
                source_branch=obj,
                service=service,
            )
            try:
                branch_diff = await merger.merge()
            except Exception as exc:
                log.exception("Merge failed, beginning rollback")
                await merger.rollback()
                raise MergeFailedError(branch_name=branch) from exc
            await merger.update_schema()

        changelog_collector = DiffChangelogCollector(diff=branch_diff, branch=obj, db=db)
        node_events = changelog_collector.collect_changelogs()
        if merger and merger.migrations:
            errors = await schema_apply_migrations(
                message=SchemaApplyMigrationData(
                    branch=merger.destination_branch,
                    new_schema=merger.destination_schema,
                    previous_schema=merger.initial_source_schema,
                    migrations=merger.migrations,
                ),
                service=service,
            )
            for error in errors:
                log.error(error)

        # -------------------------------------------------------------
        # Trigger the reconciliation of IPAM data after the merge
        # -------------------------------------------------------------
        diff_parser = await component_registry.get_component(IpamDiffParser, db=db, branch=obj)
        ipam_node_details = await diff_parser.get_changed_ipam_node_details(
            source_branch_name=obj.name,
            target_branch_name=registry.default_branch,
        )
        if ipam_node_details:
            await service.workflow.submit_workflow(
                workflow=IPAM_RECONCILIATION,
                context=context,
                parameters={"branch": registry.default_branch, "ipam_node_details": ipam_node_details},
            )
        # -------------------------------------------------------------
        # remove tracking ID from the diff because there is no diff after the merge
        # -------------------------------------------------------------
        diff_repository = await component_registry.get_component(DiffRepository, db=db, branch=obj)
        await diff_repository.mark_tracking_ids_merged(tracking_ids=[BranchTrackingId(name=obj.name)])

        # -------------------------------------------------------------
        # Generate an event to indicate that a branch has been merged
        # NOTE: we still need to convert this event and potentially pull
        #   some tasks currently executed based on the event into this workflow
        # -------------------------------------------------------------
        log_data = get_log_data()
        request_id = log_data.get("request_id", "")
        message = messages.EventBranchMerge(
            source_branch=obj.name,
            target_branch=registry.default_branch,
            context=context,
            meta=Meta(initiator_id=WORKER_IDENTITY, request_id=request_id),
        )
        await service.message_bus.send(message=message)

        events: list[InfrahubEvent] = [merge_event]

        for action, node_changelog in node_events:
            meta = EventMeta.from_parent(parent=merge_event, branch=default_branch)
            mutate_event = NodeMutatedEvent(
                kind=node_changelog.node_kind,
                node_id=node_changelog.node_id,
                data=node_changelog,
                action=MutationAction.from_diff_action(diff_action=action),
                fields=node_changelog.updated_fields,
                meta=meta,
            )
            events.append(mutate_event)

        for event in events:
            await service.event.send(event=event)


@flow(name="branch-delete", flow_run_name="Delete branch {branch}")
async def delete_branch(branch: str, context: InfrahubContext, service: InfrahubServices) -> None:
    await add_tags(branches=[branch])

    async with service.database.start_session() as db:
        obj = await Branch.get_by_name(db=db, name=str(branch))
        await obj.delete(db=db)

        event = BranchDeletedEvent(
            branch_name=branch,
            branch_id=str(obj.uuid),
            sync_with_git=obj.sync_with_git,
            meta=EventMeta.from_context(context=context, branch=registry.get_global_branch()),
        )

        await service.workflow.submit_workflow(
            workflow=BRANCH_CANCEL_PROPOSED_CHANGES, context=context, parameters={"branch_name": branch}
        )

        await service.event.send(event=event)


@flow(
    name="branch-validate",
    flow_run_name="Validate branch {branch} for conflicts",
    description="Validate if the branch has some conflicts",
    persist_result=True,
)
async def validate_branch(branch: str, service: InfrahubServices) -> State:
    await add_tags(branches=[branch])

    async with service.database.start_session() as db:
        obj = await Branch.get_by_name(db=db, name=branch)

        component_registry = get_component_registry()
        diff_repo = await component_registry.get_component(DiffRepository, db=db, branch=obj)
        has_conflicts = await diff_repo.diff_has_conflicts(
            diff_branch_name=obj.name, tracking_id=BranchTrackingId(name=obj.name)
        )
        if has_conflicts:
            return Failed(message="branch has some conflicts")
        return Completed(message="branch is valid")


@flow(name="create-branch", flow_run_name="Create branch {model.name}")
async def create_branch(model: BranchCreateModel, context: InfrahubContext, service: InfrahubServices) -> None:
    await add_tags(branches=[model.name])

    async with service.database.start_session() as db:
        try:
            await Branch.get_by_name(db=db, name=model.name)
            raise ValueError(f"The branch {model.name}, already exist")
        except BranchNotFoundError:
            pass

        data_dict: dict[str, Any] = dict(model)
        data_dict.pop("is_isolated", None)

        try:
            obj = Branch(**data_dict)
        except pydantic.ValidationError as exc:
            error_msgs = [f"invalid field {error['loc'][0]}: {error['msg']}" for error in exc.errors()]
            raise ValueError("\n".join(error_msgs)) from exc

        async with lock.registry.local_schema_lock():
            # Copy the schema from the origin branch and set the hash and the schema_changed_at value
            origin_schema = registry.schema.get_schema_branch(name=obj.origin_branch)
            new_schema = origin_schema.duplicate(name=obj.name)
            registry.schema.set_schema_branch(name=obj.name, schema=new_schema)
            obj.update_schema_hash()
            await obj.save(db=db)

            # Add Branch to registry
            registry.branch[obj.name] = obj
            await service.component.refresh_schema_hash(branches=[obj.name])

        event = BranchCreatedEvent(
            branch_name=obj.name,
            branch_id=str(obj.uuid),
            sync_with_git=obj.sync_with_git,
            meta=EventMeta.from_context(context=context, branch=registry.get_global_branch()),
        )
        await service.event.send(event=event)

        if obj.sync_with_git:
            await service.workflow.submit_workflow(
                workflow=GIT_REPOSITORIES_CREATE_BRANCH,
                context=context,
                parameters={"branch": obj.name, "branch_id": str(obj.uuid)},
            )


async def _get_diff_root(
    diff_coordinator: DiffCoordinator,
    enriched_diff_metadata: EnrichedDiffRootMetadata,
    diff_repository: DiffRepository,
    base_branch: Branch,
    target_from: Timestamp,
) -> EnrichedDiffRoot:
    default_branch_diff = await diff_coordinator.create_or_update_arbitrary_timeframe_diff(
        base_branch=base_branch,
        diff_branch=base_branch,
        from_time=target_from,
        to_time=enriched_diff_metadata.to_time,
        name=str(uuid4()),
    )
    # make sure we have the actual diff with data and not just the metadata
    if not isinstance(default_branch_diff, EnrichedDiffRoot):
        default_branch_diff = await diff_repository.get_one(
            diff_branch_name=base_branch.name, diff_id=default_branch_diff.uuid
        )

    return default_branch_diff



from typing import TYPE_CHECKING

from prefect import get_run_logger, task
from prefect.automations import AutomationCore
from prefect.cache_policies import NONE
from prefect.client.orchestration import PrefectClient
from prefect.client.schemas.filters import DeploymentFilter, DeploymentFilterName

from infrahub.trigger.models import TriggerDefinition

# from .catalogue import triggers
from .models import TriggerType

if TYPE_CHECKING:
    from uuid import UUID


@task(name="trigger-setup", task_run_name="Setup triggers of type {trigger_type.value}", cache_policy=NONE)  # type: ignore[arg-type]
async def setup_triggers(
    client: PrefectClient,
    triggers: list[TriggerDefinition],
    trigger_type: TriggerType = TriggerType.BUILTIN,
    deprecated_triggers: list[str] | None = None,
) -> None:
    log = get_run_logger()

    # -------------------------------------------------------------
    # Retrieve existing Deployments and Automation from the server
    # -------------------------------------------------------------
    deployment_names = list({name for trigger in triggers for name in trigger.get_deployment_names()})
    deployments = {
        item.name: item
        for item in await client.read_deployments(
            deployment_filter=DeploymentFilter(name=DeploymentFilterName(any_=deployment_names))
        )
    }
    deployments_mapping: dict[str, UUID] = {name: item.id for name, item in deployments.items()}
    existing_automations = {item.name: item for item in await client.read_automations()}

    trigger_automations = [
        item.name for item in await client.read_automations() if item.name.startswith(trigger_type.value)
    ]
    trigger_names = [trigger.generate_name() for trigger in triggers]

    to_delete = set(trigger_automations) - set(trigger_names)

    # -------------------------------------------------------------
    # Create or Update all triggers
    # -------------------------------------------------------------
    for trigger in triggers:
        automation = AutomationCore(
            name=trigger.generate_name(),
            description=trigger.description,
            enabled=True,
            trigger=trigger.trigger.get_prefect(),
            actions=[action.get_prefect(mapping=deployments_mapping) for action in trigger.actions],
        )

        existing_automation = existing_automations.get(trigger.name, None)

        if existing_automation:
            await client.update_automation(automation_id=existing_automation.id, automation=automation)
            log.info(f"{trigger.name} Updated")
        else:
            await client.create_automation(automation=automation)
            log.info(f"{trigger.name} Created")

    # -------------------------------------------------------------
    # Delete Triggers that shouldn't be there
    # -------------------------------------------------------------
    for item_to_delete in to_delete:
        existing_automation = existing_automations.get(item_to_delete)

        if not existing_automation:
            continue

        await client.delete_automation(automation_id=existing_automation.id)
        log.info(f"{item_to_delete} Deleted")

    # -------------------------------------------------------------
    # Delete Deprecated triggers
    # -------------------------------------------------------------
    if deprecated_triggers:
        for trigger_name in deprecated_triggers:
            existing_automation = existing_automations.get(trigger_name)

            if not existing_automation:
                continue

            await client.delete_automation(automation_id=existing_automation.id)
            log.info(f"{trigger_name} Deleted")



from infrahub.core.constants import (
    BranchSupportType,
    InfrahubKind,
)

core_transform = {
    "name": "Transformation",
    "namespace": "Core",
    "description": "Generic Transformation Object.",
    "include_in_menu": False,
    "icon": "mdi:cog-transfer",
    "label": "Transformation",
    "default_filter": "name__value",
    "order_by": ["name__value"],
    "display_labels": ["label__value"],
    "branch": BranchSupportType.AWARE.value,
    "documentation": "/topics/proposed-change",
    "uniqueness_constraints": [["name__value"]],
    "attributes": [
        {"name": "name", "kind": "Text", "unique": True},
        {"name": "label", "kind": "Text", "optional": True},
        {"name": "description", "kind": "Text", "optional": True},
        {"name": "timeout", "kind": "Number", "default_value": 10},
    ],
    "relationships": [
        {
            "name": "query",
            "peer": InfrahubKind.GRAPHQLQUERY,
            "identifier": "graphql_query__transformation",
            "kind": "Attribute",
            "cardinality": "one",
            "optional": False,
        },
        {
            "name": "repository",
            "peer": InfrahubKind.GENERICREPOSITORY,
            "kind": "Attribute",
            "cardinality": "one",
            "identifier": "repository__transformation",
            "optional": False,
        },
        {
            "name": "tags",
            "peer": InfrahubKind.TAG,
            "kind": "Attribute",
            "optional": True,
            "cardinality": "many",
        },
    ],
}

core_transform_jinja2 = {
    "name": "TransformJinja2",
    "namespace": "Core",
    "description": "A file rendered from a Jinja2 template",
    "include_in_menu": False,
    "label": "Transform Jinja2",
    "default_filter": "name__value",
    "order_by": ["name__value"],
    "display_labels": ["name__value"],
    "inherit_from": [InfrahubKind.TRANSFORM],
    "generate_profile": False,
    "branch": BranchSupportType.AWARE.value,
    "documentation": "/topics/transformation",
    "attributes": [
        {"name": "template_path", "kind": "Text"},
    ],
}

core_transform_python = {
    "name": "TransformPython",
    "namespace": "Core",
    "description": "A transform function written in Python",
    "include_in_menu": False,
    "label": "Transform Python",
    "default_filter": "name__value",
    "order_by": ["name__value"],
    "display_labels": ["name__value"],
    "inherit_from": [InfrahubKind.TRANSFORM],
    "generate_profile": False,
    "branch": BranchSupportType.AWARE.value,
    "documentation": "/topics/transformation",
    "attributes": [
        {"name": "file_path", "kind": "Text"},
        {"name": "class_name", "kind": "Text"},
    ],
}



core_object_template = {
    "name": "ObjectTemplate",
    "namespace": "Core",
    "include_in_menu": False,
    "icon": "mdi:pencil-ruler",
    "description": "Template to create pre-shaped objects.",
    "label": "Object Templates",
    "display_labels": ["template_name__value"],
    "default_filter": "template_name__value",
    "uniqueness_constraints": [["template_name__value"]],
    "attributes": [{"name": "template_name", "kind": "Text", "optional": False, "unique": True, "order_weight": 1000}],
}



lineage_owner = {
    "name": "Owner",
    "namespace": "Lineage",
    "description": "Any Entities that is responsible for some data.",
    "label": "Owner",
    "include_in_menu": False,
    "documentation": "/topics/metadata",
}

lineage_source = {
    "name": "Source",
    "namespace": "Lineage",
    "description": "Any Entities that stores or produces data.",
    "label": "Source",
    "include_in_menu": False,
    "documentation": "/topics/metadata",
}



from infrahub.core.constants import (
    BranchSupportType,
    GeneratorInstanceStatus,
    InfrahubKind,
)

core_generator_definition = {
    "name": "GeneratorDefinition",
    "namespace": "Core",
    "include_in_menu": False,
    "icon": "mdi:state-machine",
    "label": "Generator Definition",
    "default_filter": "name__value",
    "order_by": ["name__value"],
    "display_labels": ["name__value"],
    "branch": BranchSupportType.AWARE.value,
    "uniqueness_constraints": [["name__value"]],
    "generate_profile": False,
    "inherit_from": [InfrahubKind.TASKTARGET],
    "documentation": "/topics/generator",
    "attributes": [
        {"name": "name", "kind": "Text", "unique": True},
        {"name": "description", "kind": "Text", "optional": True},
        {"name": "parameters", "kind": "JSON"},
        {"name": "file_path", "kind": "Text"},
        {"name": "class_name", "kind": "Text"},
        {"name": "convert_query_response", "kind": "Boolean", "optional": True, "default_value": False},
    ],
    "relationships": [
        {
            "name": "query",
            "peer": InfrahubKind.GRAPHQLQUERY,
            "identifier": "generator_definition__graphql_query",
            "kind": "Attribute",
            "cardinality": "one",
            "optional": False,
        },
        {
            "name": "repository",
            "peer": InfrahubKind.GENERICREPOSITORY,
            "kind": "Attribute",
            "cardinality": "one",
            "identifier": "generator_definition__repository",
            "optional": False,
        },
        {
            "name": "targets",
            "peer": InfrahubKind.GENERICGROUP,
            "kind": "Attribute",
            "identifier": "generator_definition___group",
            "cardinality": "one",
            "optional": False,
        },
    ],
}

core_generator_instance = {
    "name": "GeneratorInstance",
    "namespace": "Core",
    "label": "Generator Instance",
    "include_in_menu": False,
    "icon": "mdi:file-document-outline",
    "default_filter": "name__value",
    "order_by": ["name__value"],
    "display_labels": ["name__value"],
    "branch": BranchSupportType.LOCAL.value,
    "generate_profile": False,
    "inherit_from": [InfrahubKind.TASKTARGET],
    "documentation": "/topics/generator",
    "attributes": [
        {"name": "name", "kind": "Text"},
        {
            "name": "status",
            "kind": "Text",
            "enum": GeneratorInstanceStatus.available_types(),
        },
    ],
    "relationships": [
        {
            "name": "object",
            "peer": InfrahubKind.NODE,
            "kind": "Attribute",
            "identifier": "generator__node",
            "cardinality": "one",
            "optional": False,
        },
        {
            "name": "definition",
            "peer": InfrahubKind.GENERATORDEFINITION,
            "kind": "Attribute",
            "identifier": "generator__generator_definition",
            "cardinality": "one",
            "optional": False,
        },
    ],
}



from infrahub.core.constants import (
    BranchSupportType,
    InfrahubKind,
)

core_resource_pool = {
    "name": "ResourcePool",
    "namespace": "Core",
    "label": "Resource Pool",
    "description": "The resource manager contains pools of resources to allow for automatic assignments.",
    "include_in_menu": False,
    "default_filter": "name__value",
    "order_by": ["name__value"],
    "display_labels": ["name__value"],
    "human_friendly_id": ["name__value"],
    "icon": "mdi:view-grid-outline",
    "branch": BranchSupportType.AGNOSTIC.value,
    "uniqueness_constraints": [["name__value"]],
    "generate_profile": False,
    "attributes": [
        {
            "name": "name",
            "kind": "Text",
            "order_weight": 1000,
            "unique": True,
        },
        {
            "name": "description",
            "kind": "Text",
            "optional": True,
            "order_weight": 2000,
        },
    ],
}

core_ip_prefix_pool = {
    "name": "IPPrefixPool",
    "namespace": "Core",
    "description": "A pool of IP prefix resources",
    "label": "IP Prefix Pool",
    "include_in_menu": False,
    "branch": BranchSupportType.AGNOSTIC.value,
    "generate_profile": False,
    "inherit_from": [InfrahubKind.RESOURCEPOOL, InfrahubKind.LINEAGESOURCE],
    "attributes": [
        {
            "name": "default_prefix_length",
            "kind": "Number",
            "description": "The default prefix length as an integer for prefixes allocated from this pool.",
            "optional": True,
            "order_weight": 5000,
        },
        {
            "name": "default_member_type",
            "kind": "Text",
            "enum": ["prefix", "address"],
            "default_value": "prefix",
            "optional": True,
            "order_weight": 3000,
        },
        {
            "name": "default_prefix_type",
            "kind": "Text",
            "optional": True,
            "order_weight": 4000,
        },
    ],
    "relationships": [
        {
            "name": "resources",
            "peer": "BuiltinIPPrefix",
            "kind": "Attribute",
            "identifier": "prefixpool__resource",
            "cardinality": "many",
            "branch": BranchSupportType.AGNOSTIC.value,
            "optional": False,
            "order_weight": 6000,
        },
        {
            "name": "ip_namespace",
            "peer": "BuiltinIPNamespace",
            "kind": "Attribute",
            "identifier": "prefixpool__ipnamespace",
            "cardinality": "one",
            "branch": BranchSupportType.AGNOSTIC.value,
            "optional": False,
            "order_weight": 7000,
        },
    ],
}

core_ip_address_pool = {
    "name": "IPAddressPool",
    "namespace": "Core",
    "description": "A pool of IP address resources",
    "label": "IP Address Pool",
    "include_in_menu": False,
    "branch": BranchSupportType.AGNOSTIC.value,
    "generate_profile": False,
    "inherit_from": [InfrahubKind.RESOURCEPOOL, InfrahubKind.LINEAGESOURCE],
    "attributes": [
        {
            "name": "default_address_type",
            "kind": "Text",
            "optional": False,
            "description": "The object type to create when reserving a resource in the pool",
            "order_weight": 3000,
        },
        {
            "name": "default_prefix_length",
            "kind": "Number",
            "description": "The default prefix length as an integer for addresses allocated from this pool.",
            "optional": True,
            "order_weight": 4000,
        },
    ],
    "relationships": [
        {
            "name": "resources",
            "peer": "BuiltinIPPrefix",
            "kind": "Attribute",
            "identifier": "ipaddresspool__resource",
            "cardinality": "many",
            "branch": BranchSupportType.AGNOSTIC.value,
            "optional": False,
            "order_weight": 5000,
        },
        {
            "name": "ip_namespace",
            "peer": "BuiltinIPNamespace",
            "kind": "Attribute",
            "identifier": "ipaddresspool__ipnamespace",
            "cardinality": "one",
            "branch": BranchSupportType.AGNOSTIC.value,
            "optional": False,
            "order_weight": 6000,
        },
    ],
}

core_number_pool = {
    "name": "NumberPool",
    "namespace": "Core",
    "description": "A pool of number resources",
    "label": "Number Pool",
    "include_in_menu": False,
    "branch": BranchSupportType.AGNOSTIC.value,
    "generate_profile": False,
    "inherit_from": [InfrahubKind.RESOURCEPOOL, InfrahubKind.LINEAGESOURCE],
    "attributes": [
        {
            "name": "node",
            "kind": "Text",
            "optional": False,
            "description": "The model of the object that requires integers to be allocated",
            "order_weight": 3000,
        },
        {
            "name": "node_attribute",
            "kind": "Text",
            "description": "The attribute of the selected model",
            "optional": False,
            "order_weight": 4000,
        },
        {
            "name": "start_range",
            "kind": "Number",
            "optional": False,
            "description": "The start range for the pool",
            "order_weight": 5000,
        },
        {
            "name": "end_range",
            "kind": "Number",
            "optional": False,
            "description": "The end range for the pool",
            "order_weight": 6000,
        },
    ],
}



from infrahub.core.constants import (
    AccountStatus,
    AccountType,
    BranchSupportType,
    InfrahubKind,
)

core_account = {
    "name": "Account",
    "namespace": "Core",
    "description": "User Account for Infrahub",
    "include_in_menu": False,
    "label": "Account",
    "icon": "mdi:account",
    "default_filter": "name__value",
    "order_by": ["name__value"],
    "display_labels": ["label__value"],
    "generate_profile": False,
    "branch": BranchSupportType.AGNOSTIC.value,
    "inherit_from": [InfrahubKind.LINEAGEOWNER, InfrahubKind.LINEAGESOURCE, InfrahubKind.GENERICACCOUNT],
}

core_account_token = {
    "name": "AccountToken",
    "namespace": "Internal",
    "description": "Token for User Account",
    "include_in_menu": False,
    "label": "Account Token",
    "default_filter": "token__value",
    "display_labels": ["token__value"],
    "generate_profile": False,
    "branch": BranchSupportType.AGNOSTIC.value,
    "uniqueness_constraints": [["token__value"]],
    "documentation": "/topics/auth",
    "attributes": [
        {"name": "name", "kind": "Text", "optional": True},
        {"name": "token", "kind": "Text", "unique": True},
        {"name": "expiration", "kind": "DateTime", "optional": True},
    ],
    "relationships": [
        {
            "name": "account",
            "peer": InfrahubKind.GENERICACCOUNT,
            "optional": False,
            "cardinality": "one",
            "identifier": "account__token",
        },
    ],
}

core_password_credential = {
    "name": "PasswordCredential",
    "namespace": "Core",
    "description": "Username/Password based credential",
    "include_in_menu": False,
    "label": "Username / Password",
    "generate_profile": False,
    "branch": BranchSupportType.AGNOSTIC.value,
    "inherit_from": [InfrahubKind.CREDENTIAL],
    "attributes": [
        {
            "name": "username",
            "kind": "Text",
            "optional": True,
            "branch": BranchSupportType.AGNOSTIC.value,
            "order_weight": 6000,
        },
        {
            "name": "password",
            "kind": "Password",
            "optional": True,
            "branch": BranchSupportType.AGNOSTIC.value,
            "order_weight": 7000,
        },
    ],
}

core_refresh_token = {
    "name": "RefreshToken",
    "namespace": "Internal",
    "description": "Refresh Token",
    "include_in_menu": False,
    "label": "Refresh Token",
    "display_labels": [],
    "generate_profile": False,
    "branch": BranchSupportType.AGNOSTIC.value,
    "attributes": [
        {"name": "expiration", "kind": "DateTime", "optional": False},
    ],
    "relationships": [
        {
            "name": "account",
            "peer": InfrahubKind.GENERICACCOUNT,
            "optional": False,
            "cardinality": "one",
            "identifier": "account__refreshtoken",
        },
    ],
}

core_credential = {
    "name": "Credential",
    "namespace": "Core",
    "description": "A credential that could be referenced to access external services.",
    "include_in_menu": False,
    "label": "Credential",
    "default_filter": "name__value",
    "order_by": ["name__value"],
    "display_labels": ["label__value"],
    "icon": "mdi:key-variant",
    "human_friendly_id": ["name__value"],
    "branch": BranchSupportType.AGNOSTIC.value,
    "uniqueness_constraints": [["name__value"]],
    "documentation": "/topics/auth",
    "attributes": [
        {"name": "name", "kind": "Text", "unique": True, "order_weight": 1000},
        {"name": "label", "kind": "Text", "optional": True, "order_weight": 2000},
        {"name": "description", "kind": "Text", "optional": True, "order_weight": 3000},
    ],
}

core_generic_account = {
    "name": "GenericAccount",
    "namespace": "Core",
    "description": "User Account for Infrahub",
    "include_in_menu": False,
    "label": "Account",
    "icon": "mdi:account",
    "default_filter": "name__value",
    "order_by": ["name__value"],
    "display_labels": ["label__value"],
    "human_friendly_id": ["name__value"],
    "branch": BranchSupportType.AGNOSTIC.value,
    "documentation": "/topics/auth",
    "uniqueness_constraints": [["name__value"]],
    "attributes": [
        {"name": "name", "kind": "Text", "unique": True},
        {"name": "password", "kind": "HashedPassword", "unique": False},
        {"name": "label", "kind": "Text", "optional": True},
        {"name": "description", "kind": "Text", "optional": True},
        {
            "name": "account_type",
            "kind": "Text",
            "default_value": AccountType.USER.value,
            "enum": AccountType.available_types(),
        },
        {
            "name": "status",
            "kind": "Dropdown",
            "choices": [
                {
                    "name": AccountStatus.ACTIVE.value,
                    "label": "Active",
                    "description": "Account is allowed to login",
                    "color": "#52be80",
                },
                {
                    "name": AccountStatus.INACTIVE.value,
                    "label": "Inactive",
                    "description": "Account is not allowed to login",
                    "color": "#e74c3c",
                },
            ],
            "default_value": AccountStatus.ACTIVE.value,
        },
    ],
    "relationships": [{"name": "tokens", "peer": InfrahubKind.ACCOUNTTOKEN, "optional": True, "cardinality": "many"}],
}



from __future__ import annotations

from typing import TYPE_CHECKING, Any

from graphene import Boolean, DateTime, Field, InputObjectType, Int, Interface, List, NonNull, ObjectType, String
from graphene.types.generic import GenericScalar

from .common import RelatedNode
from .enums import DiffAction

if TYPE_CHECKING:
    from graphql import GraphQLResolveInfo


class InfrahubMutatedAttribute(ObjectType):
    name = String(required=True)
    action = DiffAction(required=True)
    value = String(required=False)
    kind = String(required=True)
    value_previous = String(required=False)


class EventNodeInterface(Interface):
    id = String(required=True, description="The ID of the event.")
    event = String(required=True, description="The name of the event.")
    branch = String(required=False, description="The branch where the event occurred.")
    account_id = String(required=False, description="The account ID that triggered the event.")
    occurred_at = DateTime(required=True, description="The timestamp when the event occurred.")
    level = Int(
        required=True,
        description="The level of the event 0 is a root level event, the child events will have 1 and grand children 2.",
    )
    primary_node = Field(
        RelatedNode, required=False, description="The primary Infrahub node this event is associated with."
    )
    related_nodes = List(
        NonNull(RelatedNode), required=True, description="Related Infrahub nodes this event is associated with."
    )
    has_children = Boolean(
        required=True, description="Indicates if the event is expected to have child events under it"
    )
    parent_id = String(required=False, description="The event ID of the direct parent to this event.")

    @classmethod
    def resolve_type(
        cls,
        instance: dict[str, Any],
        info: GraphQLResolveInfo,  # noqa: ARG003
    ) -> type[ObjectType]:
        if "event" in instance:
            return EVENT_TYPES.get(instance["event"], StandardEvent)
        return StandardEvent


class EventNodes(ObjectType):
    node = Field(EventNodeInterface)


class BranchEventTypeFilter(InputObjectType):
    branches = List(NonNull(String), required=True, description="Name of impacted branches")


class EventTypeFilter(InputObjectType):
    branch_merged = Field(
        BranchEventTypeFilter, required=False, description="Filters specific to infrahub.branch.merged events"
    )
    branch_rebased = Field(
        BranchEventTypeFilter, required=False, description="Filters specific to infrahub.branch.rebased events"
    )


# ---------------------------------------
# Branch events
# ---------------------------------------
class BranchCreatedEvent(ObjectType):
    class Meta:
        interfaces = (EventNodeInterface,)

    created_branch = String(required=True, description="The name of the branch that was created")
    payload = Field(GenericScalar, required=True)


class BranchMergedEvent(ObjectType):
    class Meta:
        interfaces = (EventNodeInterface,)

    source_branch = String(required=True, description="The name of the branch that was merged into the default branch")


class BranchRebasedEvent(ObjectType):
    class Meta:
        interfaces = (EventNodeInterface,)

    rebased_branch = String(
        required=True, description="The name of the branch that was rebased and aligned with the default branch"
    )
    payload = Field(GenericScalar, required=True)


class BranchDeletedEvent(ObjectType):
    class Meta:
        interfaces = (EventNodeInterface,)

    deleted_branch = String(required=True, description="The name of the branch that was deleted")
    payload = Field(GenericScalar, required=True)


# ---------------------------------------
# Node/Object events
# ---------------------------------------
class NodeMutatedEvent(ObjectType):
    class Meta:
        interfaces = (EventNodeInterface,)

    payload = Field(GenericScalar, required=True)
    attributes = Field(List(of_type=NonNull(InfrahubMutatedAttribute), required=True), required=True)


class ArtifactEvent(ObjectType):
    class Meta:
        interfaces = (EventNodeInterface,)

    checksum = String(required=True, description="The current checksum of the artifact")
    checksum_previous = String(required=False, description="The previous checksum of the artifact")
    storage_id = String(required=True, description="The current storage_id of the artifact")
    storage_id_previous = String(required=False, description="The previous storage_id of the artifact")
    artifact_definition_id = String(required=True, description="Artifact definition ID")


class GroupEvent(ObjectType):
    class Meta:
        interfaces = (EventNodeInterface,)

    members = List(NonNull(RelatedNode), required=True, description="Group members modified in this event")
    ancestors = List(NonNull(RelatedNode), required=True, description="Ancestor groups of this impacted group")


class StandardEvent(ObjectType):
    class Meta:
        interfaces = (EventNodeInterface,)

    payload = Field(GenericScalar, required=True)


EVENT_TYPES: dict[str, type[ObjectType]] = {
    "infrahub.artifact.created": ArtifactEvent,
    "infrahub.artifact.updated": ArtifactEvent,
    "infrahub.node.created": NodeMutatedEvent,
    "infrahub.node.updated": NodeMutatedEvent,
    "infrahub.node.deleted": NodeMutatedEvent,
    "infrahub.branch.created": BranchCreatedEvent,
    "infrahub.branch.merged": BranchMergedEvent,
    "infrahub.branch.rebased": BranchRebasedEvent,
    "infrahub.branch.deleted": BranchDeletedEvent,
    "infrahub.group.member_added": GroupEvent,
    "infrahub.group.member_removed": GroupEvent,
    "undefined": StandardEvent,
}



DEPRECATED_STATIC_TRIGGER_NAMES = [
    "Trigger-branch-create-events",  # used in 1.1.x
    "Trigger-branch-remove-events",  # used in 1.1.x
    "Trigger-webhook-update-configuration",  # used in 1.1.x
    "Trigger-repository-commit-update-event",  # used in 1.1.x
    "Trigger-schema-update-event",  # used in 1.1.x
]

NAME_SEPARATOR = "::"



from infrahub.core.constants import (
    AllowOverrideType,
    BranchSupportType,
    InfrahubKind,
    RelationshipDeleteBehavior,
)

builtin_ipam = {
    "name": "IPNamespace",
    "namespace": "Builtin",
    "label": "IP Namespace",
    "description": "A generic container for IP prefixes and IP addresses",
    "include_in_menu": False,
    "default_filter": "name__value",
    "order_by": ["name__value"],
    "display_labels": ["name__value"],
    "icon": "mdi:format-list-group",
    "branch": BranchSupportType.AWARE.value,
    "uniqueness_constraints": [["name__value"]],
    "generate_profile": False,
    "attributes": [
        {
            "name": "name",
            "kind": "Text",
            "unique": True,
            "branch": BranchSupportType.AWARE.value,
            "order_weight": 1000,
        },
        {
            "name": "description",
            "kind": "Text",
            "optional": True,
            "branch": BranchSupportType.AWARE.value,
            "order_weight": 2000,
        },
    ],
    "relationships": [
        {
            "name": "ip_prefixes",
            "label": "IP Prefixes",
            "peer": InfrahubKind.IPPREFIX,
            "identifier": "ip_namespace__ip_prefix",
            "optional": True,
            "cardinality": "many",
            "on_delete": RelationshipDeleteBehavior.CASCADE,
            "allow_override": AllowOverrideType.NONE,
        },
        {
            "name": "ip_addresses",
            "label": "IP Addresses",
            "peer": InfrahubKind.IPADDRESS,
            "identifier": "ip_namespace__ip_address",
            "optional": True,
            "cardinality": "many",
            "on_delete": RelationshipDeleteBehavior.CASCADE,
            "allow_override": AllowOverrideType.NONE,
        },
    ],
}

builtin_ip_prefix = {
    "name": "IPPrefix",
    "label": "IP Prefix",
    "namespace": "Builtin",
    "description": "IPv6 or IPv4 prefix also referred as network",
    "include_in_menu": False,
    "default_filter": "prefix__value",
    "order_by": ["prefix__version", "prefix__binary_address", "prefix__prefixlen"],
    "display_labels": ["prefix__value"],
    "icon": "mdi:ip-network",
    "branch": BranchSupportType.AWARE.value,
    "hierarchical": True,
    "attributes": [
        {
            "name": "prefix",
            "kind": "IPNetwork",
            "branch": BranchSupportType.AWARE.value,
            "order_weight": 1000,
        },
        {
            "name": "description",
            "kind": "Text",
            "optional": True,
            "branch": BranchSupportType.AWARE.value,
            "order_weight": 2000,
        },
        {
            "name": "member_type",
            "kind": "Dropdown",
            "choices": [
                {
                    "name": "prefix",
                    "label": "Prefix",
                    "description": "Prefix serves as container for other prefixes",
                },
                {
                    "name": "address",
                    "label": "Address",
                    "description": "Prefix serves as subnet for IP addresses",
                },
            ],
            "branch": BranchSupportType.AWARE.value,
            "default_value": "address",
            "order_weight": 3000,
        },
        {
            "name": "is_pool",
            "kind": "Boolean",
            "branch": BranchSupportType.AWARE.value,
            "default_value": False,
            "order_weight": 4000,
            "description": "All IP addresses within this prefix are considered usable",
        },
        {
            "name": "is_top_level",
            "kind": "Boolean",
            "read_only": True,
            "optional": True,
            "allow_override": AllowOverrideType.NONE,
        },
        {
            "name": "utilization",
            "kind": "Number",
            "read_only": True,
            "optional": True,
            "allow_override": AllowOverrideType.NONE,
        },
        {
            "name": "netmask",
            "kind": "Text",
            "read_only": True,
            "optional": True,
            "allow_override": AllowOverrideType.NONE,
        },
        {
            "name": "hostmask",
            "kind": "Text",
            "read_only": True,
            "optional": True,
            "allow_override": AllowOverrideType.NONE,
        },
        {
            "name": "network_address",
            "kind": "Text",
            "read_only": True,
            "optional": True,
            "allow_override": AllowOverrideType.NONE,
        },
        {
            "name": "broadcast_address",
            "kind": "Text",
            "read_only": True,
            "optional": True,
            "allow_override": AllowOverrideType.NONE,
        },
    ],
    "relationships": [
        {
            "name": "ip_namespace",
            "label": "IP Namespace",
            "peer": InfrahubKind.IPNAMESPACE,
            "identifier": "ip_namespace__ip_prefix",
            "optional": True,
            "cardinality": "one",
            "allow_override": AllowOverrideType.NONE,
        },
        {
            "name": "ip_addresses",
            "label": "IP Addresses",
            "peer": InfrahubKind.IPADDRESS,
            "identifier": "ip_prefix__ip_address",
            "optional": True,
            "cardinality": "many",
            "allow_override": AllowOverrideType.NONE,
            "read_only": True,
        },
        {
            "name": "resource_pool",
            "peer": "CoreIPAddressPool",
            "identifier": "ipaddresspool__resource",
            "cardinality": "many",
            "branch": BranchSupportType.AGNOSTIC.value,
            "optional": True,
            "read_only": True,
        },
    ],
}

builtin_ip_address = {
    "name": "IPAddress",
    "label": "IP Address",
    "namespace": "Builtin",
    "description": "IPv6 or IPv4 address",
    "include_in_menu": False,
    "default_filter": "address__value",
    "order_by": ["address__version", "address__binary_address"],
    "display_labels": ["address__value"],
    "icon": "mdi:ip-outline",
    "branch": BranchSupportType.AWARE.value,
    "attributes": [
        {
            "name": "address",
            "kind": "IPHost",
            "branch": BranchSupportType.AWARE.value,
            "order_weight": 1000,
        },
        {
            "name": "description",
            "kind": "Text",
            "optional": True,
            "branch": BranchSupportType.AWARE.value,
            "order_weight": 2000,
        },
    ],
    "relationships": [
        {
            "name": "ip_namespace",
            "label": "IP Namespace",
            "peer": InfrahubKind.IPNAMESPACE,
            "identifier": "ip_namespace__ip_address",
            "optional": True,
            "cardinality": "one",
            "allow_override": AllowOverrideType.NONE,
        },
        {
            "name": "ip_prefix",
            "label": "IP Prefix",
            "peer": InfrahubKind.IPPREFIX,
            "identifier": "ip_prefix__ip_address",
            "optional": True,
            "cardinality": "one",
            "allow_override": AllowOverrideType.NONE,
            "read_only": True,
        },
    ],
}

core_ipam_namespace = {
    "name": "Namespace",
    "namespace": "Ipam",
    "description": "A namespace that segments IPAM",
    "label": "IPAM Namespace",
    "default_filter": "name__value",
    "human_friendly_id": ["name__value"],
    "order_by": ["name__value"],
    "display_labels": ["name__value"],
    "include_in_menu": False,
    "icon": "mdi:format-list-group",
    "branch": BranchSupportType.AWARE.value,
    "inherit_from": [InfrahubKind.IPNAMESPACE],
    "attributes": [{"name": "default", "kind": "Boolean", "optional": True, "read_only": True, "order_weight": 9000}],
}



from infrahub.trigger.models import BuiltinTriggerDefinition, EventTrigger, ExecuteWorkflow
from infrahub.workflows.catalogue import (
    COMPUTED_ATTRIBUTE_REMOVE_PYTHON,
    COMPUTED_ATTRIBUTE_SETUP,
    COMPUTED_ATTRIBUTE_SETUP_PYTHON,
)

TRIGGER_COMPUTED_ATTRIBUTE_PYTHON_SETUP_BRANCH = BuiltinTriggerDefinition(
    name="computed-attribute-python-setup-on-branch-creation",
    previous_names={"Trigger-schema-update-event"},
    description="Trigger actions on branch create event",
    trigger=EventTrigger(events={"infrahub.branch.created"}),
    actions=[
        ExecuteWorkflow(
            workflow=COMPUTED_ATTRIBUTE_SETUP_PYTHON,
            parameters={
                "branch_name": "{{ event.resource['infrahub.branch.name'] }}",
                "trigger_updates": False,
                "context": {
                    "__prefect_kind": "json",
                    "value": {"__prefect_kind": "jinja", "template": "{{ event.payload['context'] | tojson }}"},
                },
            },
        )
    ],
)

TRIGGER_COMPUTED_ATTRIBUTE_PYTHON_SETUP_COMMIT = BuiltinTriggerDefinition(
    name="computed-attribute-python-setup-on-commit",
    description="Trigger actions on branch create event",
    trigger=EventTrigger(events={"infrahub.repository.update_commit"}),
    actions=[
        ExecuteWorkflow(
            workflow=COMPUTED_ATTRIBUTE_SETUP_PYTHON,
            parameters={
                "branch_name": "{{ event.resource['infrahub.branch.name'] }}",
                "commit": "{{ event.payload['commit'] }}",
                "context": {
                    "__prefect_kind": "json",
                    "value": {"__prefect_kind": "jinja", "template": "{{ event.payload['context'] | tojson }}"},
                },
            },
        )
    ],
)

TRIGGER_COMPUTED_ATTRIBUTE_PYTHON_CLEAN_BRANCH = BuiltinTriggerDefinition(
    name="computed-attribute-python-cleanup-on-branch-deletion",
    description="Trigger actions on branch delete event",
    trigger=EventTrigger(events={"infrahub.branch.deleted"}),
    actions=[
        ExecuteWorkflow(
            workflow=COMPUTED_ATTRIBUTE_REMOVE_PYTHON,
            parameters={
                "branch_name": "{{ event.resource['infrahub.branch.name'] }}",
                "context": {
                    "__prefect_kind": "json",
                    "value": {"__prefect_kind": "jinja", "template": "{{ event.payload['context'] | tojson }}"},
                },
            },
        )
    ],
)

TRIGGER_COMPUTED_ATTRIBUTE_ALL_SCHEMA = BuiltinTriggerDefinition(
    name="computed-attribute-all-setup-on-schema-update",
    trigger=EventTrigger(events={"infrahub.schema.update"}),
    actions=[
        ExecuteWorkflow(
            workflow=COMPUTED_ATTRIBUTE_SETUP,
            parameters={
                "branch_name": "{{ event.resource['infrahub.branch.name'] }}",
                "context": {
                    "__prefect_kind": "json",
                    "value": {"__prefect_kind": "jinja", "template": "{{ event.payload['context'] | tojson }}"},
                },
            },
        ),
        ExecuteWorkflow(
            workflow=COMPUTED_ATTRIBUTE_SETUP_PYTHON,
            parameters={
                "branch_name": "{{ event.resource['infrahub.branch.name'] }}",
                "context": {
                    "__prefect_kind": "json",
                    "value": {"__prefect_kind": "jinja", "template": "{{ event.payload['context'] | tojson }}"},
                },
            },
        ),
    ],
)



from infrahub.core.constants import (
    BranchSupportType,
    InfrahubKind,
    RelationshipDeleteBehavior,
)

core_propose_change_comment = {
    "name": "Comment",
    "namespace": "Core",
    "description": "A comment on a Proposed Change",
    "label": "Comment",
    "display_labels": ["text__value"],
    "order_by": ["created_at__value"],
    "include_in_menu": False,
    "branch": BranchSupportType.AGNOSTIC.value,
    "attributes": [
        {"name": "text", "kind": "TextArea", "unique": False, "optional": False},
        {"name": "created_at", "kind": "DateTime", "optional": True},
    ],
    "relationships": [
        {
            "name": "created_by",
            "peer": InfrahubKind.GENERICACCOUNT,
            "optional": True,
            "branch": BranchSupportType.AGNOSTIC.value,
            "cardinality": "one",
            "identifier": "comment__account",
        },
    ],
}

core_thread = {
    "name": "Thread",
    "namespace": "Core",
    "description": "A thread on a Proposed Change",
    "label": "Thread",
    "order_by": ["created_at__value"],
    "branch": BranchSupportType.AGNOSTIC.value,
    "include_in_menu": False,
    "attributes": [
        {"name": "label", "kind": "Text", "optional": True},
        {"name": "resolved", "kind": "Boolean", "default_value": False},
        {"name": "created_at", "kind": "DateTime", "optional": True},
    ],
    "relationships": [
        {
            "name": "change",
            "peer": InfrahubKind.PROPOSEDCHANGE,
            "identifier": "proposedchange__thread",
            "kind": "Parent",
            "optional": False,
            "cardinality": "one",
        },
        {
            "name": "comments",
            "peer": InfrahubKind.THREADCOMMENT,
            "identifier": "thread__threadcomment",
            "kind": "Component",
            "optional": True,
            "cardinality": "many",
            "on_delete": RelationshipDeleteBehavior.CASCADE,
        },
        {
            "name": "created_by",
            "peer": InfrahubKind.GENERICACCOUNT,
            "identifier": "thread__account",
            "optional": True,
            "branch": BranchSupportType.AGNOSTIC.value,
            "cardinality": "one",
        },
    ],
}

core_change_thread = {
    "name": "ChangeThread",
    "namespace": "Core",
    "description": "A thread on proposed change",
    "include_in_menu": False,
    "label": "Change Thread",
    "branch": BranchSupportType.AGNOSTIC.value,
    "inherit_from": [InfrahubKind.THREAD],
    "generate_profile": False,
    "attributes": [],
    "relationships": [],
}

core_file_thread = {
    "name": "FileThread",
    "namespace": "Core",
    "description": "A thread related to a file on a proposed change",
    "include_in_menu": False,
    "label": "Thread - File",
    "branch": BranchSupportType.AGNOSTIC.value,
    "inherit_from": [InfrahubKind.THREAD],
    "generate_profile": False,
    "attributes": [
        {"name": "file", "kind": "Text", "optional": True},
        {"name": "commit", "kind": "Text", "optional": True},
        {"name": "line_number", "kind": "Number", "optional": True},
    ],
    "relationships": [
        {
            "name": "repository",
            "peer": InfrahubKind.REPOSITORY,
            "optional": False,
            "cardinality": "one",
            "branch": BranchSupportType.AGNOSTIC.value,
        },
    ],
}

core_artifact_thread = {
    "name": "ArtifactThread",
    "namespace": "Core",
    "description": "A thread related to an artifact on a proposed change",
    "include_in_menu": False,
    "label": "Thread - Artifact",
    "branch": BranchSupportType.AGNOSTIC.value,
    "inherit_from": [InfrahubKind.THREAD],
    "generate_profile": False,
    "attributes": [
        {"name": "artifact_id", "kind": "Text", "optional": True},
        {"name": "storage_id", "kind": "Text", "optional": True},
        {"name": "line_number", "kind": "Number", "optional": True},
    ],
    "relationships": [],
}

core_object_thread = {
    "name": "ObjectThread",
    "namespace": "Core",
    "description": "A thread related to an object on a proposed change",
    "include_in_menu": False,
    "label": "Thread - Object",
    "branch": BranchSupportType.AGNOSTIC.value,
    "inherit_from": [InfrahubKind.THREAD],
    "generate_profile": False,
    "attributes": [
        {"name": "object_path", "kind": "Text", "optional": False},
    ],
    "relationships": [],
}

core_change_comment = {
    "name": "ChangeComment",
    "namespace": "Core",
    "description": "A comment on proposed change",
    "include_in_menu": False,
    "label": "Change Comment",
    "default_filter": "text__value",
    "display_labels": ["text__value"],
    "branch": BranchSupportType.AGNOSTIC.value,
    "inherit_from": [InfrahubKind.COMMENT],
    "generate_profile": False,
    "relationships": [
        {
            "name": "change",
            "kind": "Parent",
            "peer": InfrahubKind.PROPOSEDCHANGE,
            "cardinality": "one",
            "optional": False,
        },
    ],
}

core_thread_comment = {
    "name": "ThreadComment",
    "namespace": "Core",
    "description": "A comment on thread within a Proposed Change",
    "include_in_menu": False,
    "label": "Thread Comment",
    "default_filter": "text__value",
    "display_labels": ["text__value"],
    "branch": BranchSupportType.AGNOSTIC.value,
    "inherit_from": [InfrahubKind.COMMENT],
    "generate_profile": False,
    "attributes": [],
    "relationships": [
        {
            "name": "thread",
            "peer": InfrahubKind.THREAD,
            "kind": "Parent",
            "identifier": "thread__threadcomment",
            "cardinality": "one",
            "optional": False,
        },
    ],
}



from __future__ import annotations

import base64
import hashlib
import hmac
from typing import TYPE_CHECKING, Any, Optional, Union
from uuid import UUID, uuid4

from pydantic import BaseModel, ConfigDict, Field, computed_field
from typing_extensions import Self

from infrahub.core import registry
from infrahub.core.constants import InfrahubKind
from infrahub.core.timestamp import Timestamp
from infrahub.git.repository import InfrahubReadOnlyRepository, InfrahubRepository
from infrahub.trigger.models import EventTrigger, ExecuteWorkflow, TriggerDefinition, TriggerType
from infrahub.workflows.catalogue import WEBHOOK_PROCESS

if TYPE_CHECKING:
    from httpx import Response
    from infrahub_sdk.protocols import CoreCustomWebhook, CoreStandardWebhook, CoreTransformPython, CoreWebhook

    from infrahub.services import InfrahubServices


class WebhookTriggerDefinition(TriggerDefinition):
    type: TriggerType = TriggerType.WEBHOOK

    @classmethod
    def from_object(cls, obj: CoreWebhook) -> Self:
        event_trigger = EventTrigger()
        if obj.event_type.value == "all":
            event_trigger.events.add("infrahub.*")
        else:
            event_trigger.events.add(obj.event_type.value)

        if obj.branch_scope.value == "default_branch":
            event_trigger.match_related = {
                "prefect.resource.role": "infrahub.branch",
                "infrahub.resource.label": registry.default_branch,
            }
        elif obj.branch_scope.value == "other_branches":
            event_trigger.match_related = {
                "prefect.resource.role": "infrahub.branch",
                "infrahub.resource.label": f"!{registry.default_branch}",
            }

        definition = cls(
            name=obj.name.value,
            trigger=event_trigger,
            actions=[
                ExecuteWorkflow(
                    workflow=WEBHOOK_PROCESS,
                    parameters={
                        "webhook_id": obj.id,
                        "webhook_kind": obj.get_kind(),
                        "event_id": "{{ event.id }}",
                        "event_type": "{{ event.event }}",
                        "event_occured_at": "{{ event.occurred }}",
                        "event_payload": {
                            "__prefect_kind": "json",
                            "value": {"__prefect_kind": "jinja", "template": "{{ event.payload | tojson }}"},
                        },
                    },
                ),
            ],
        )

        return definition


class EventContext(BaseModel):
    id: str = Field(..., description="The internal id of the event")
    branch: str | None = Field(None, description="The branch associated with the event")
    account_id: str | None = Field(None, description="The id of the account associated with the event")
    occured_at: str = Field(..., description="The time when the event occurred")
    event: str = Field(..., description="The event type")

    @classmethod
    def from_event(cls, event_id: str, event_type: str, event_occured_at: str, event_payload: dict[str, Any]) -> Self:
        """Extract the context from the raw event we are getting from Prefect."""

        infrahub_context: dict[str, Any] = event_payload.get("context", {})
        account_info: dict[str, Any] = infrahub_context.get("account", {})
        branch_info: dict[str, Any] = infrahub_context.get("branch", {})

        return cls(
            id=event_id,
            branch=branch_info.get("name")
            if branch_info and branch_info.get("name") != registry.get_global_branch().name
            else None,
            account_id=account_info.get("account_id"),
            occured_at=event_occured_at,
            event=event_type,
        )


class Webhook(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True)
    name: str = Field(...)
    url: str = Field(...)
    event_type: str = Field(...)
    validate_certificates: bool = Field(...)
    _payload: Any = None
    _headers: Optional[dict[str, Any]] = None

    async def _prepare_payload(self, data: dict[str, Any], context: EventContext, service: InfrahubServices) -> None:  # noqa: ARG002
        self._payload = {"data": data, **context.model_dump()}

    def _assign_headers(self) -> None:
        self._headers = {}

    @computed_field  # type: ignore[prop-decorator]
    @property
    def webhook_type(self) -> str:
        return self.__class__.__name__

    async def prepare(self, data: dict[str, Any], context: EventContext, service: InfrahubServices) -> None:
        await self._prepare_payload(data=data, context=context, service=service)
        self._assign_headers()

    async def send(self, data: dict[str, Any], context: EventContext, service: InfrahubServices) -> Response:
        await self.prepare(data=data, context=context, service=service)
        return await service.http.post(url=self.url, json=self.get_payload(), headers=self._headers)

    def get_payload(self) -> dict[str, Any]:
        return self._payload

    def to_cache(self) -> dict[str, Any]:
        return self.model_dump()

    @classmethod
    def from_cache(cls, data: dict[str, Any]) -> Self:
        return cls(**data)


class CustomWebhook(Webhook):
    """Custom webhook"""

    @classmethod
    def from_object(cls, obj: CoreCustomWebhook) -> Self:
        return cls(
            name=obj.name.value,
            url=obj.url.value,
            event_type=obj.event_type.value,
            validate_certificates=obj.validate_certificates.value or False,
        )


class StandardWebhook(Webhook):
    shared_key: str = Field(...)

    def _assign_headers(self, uuid: UUID | None = None, at: Timestamp | None = None) -> None:
        message_id = f"msg_{uuid.hex}" if uuid else f"msg_{uuid4().hex}"
        timestamp = str(at.to_timestamp()) if at else str(Timestamp().to_timestamp())
        payload = self._payload or {}
        unsigned_data = f"{message_id}.{timestamp}.{payload}".encode()
        signature = self._sign(data=unsigned_data)

        self._headers = {
            "Accept": "application/json",
            "Content-Type": "application/json",
            "webhook-id": message_id,
            "webhook-timestamp": timestamp,
            "webhook-signature": f"v1,{base64.b64encode(signature).decode('utf-8')}",
        }

    def _sign(self, data: bytes) -> bytes:
        return hmac.new(key=self.shared_key.encode(), msg=data, digestmod=hashlib.sha256).digest()

    @classmethod
    def from_object(cls, obj: CoreStandardWebhook) -> Self:
        return cls(
            name=obj.name.value,
            url=obj.url.value,
            event_type=obj.event_type.value,
            validate_certificates=obj.validate_certificates.value or False,
            shared_key=obj.shared_key.value,
        )


class TransformWebhook(Webhook):
    repository_id: str = Field(...)
    repository_name: str = Field(...)
    repository_kind: str = Field(...)
    transform_name: str = Field(...)
    transform_class: str = Field(...)
    transform_file: str = Field(...)
    transform_timeout: int = Field(...)

    async def _prepare_payload(self, data: dict[str, Any], context: EventContext, service: InfrahubServices) -> None:
        repo: Union[InfrahubReadOnlyRepository, InfrahubRepository]
        if self.repository_kind == InfrahubKind.READONLYREPOSITORY:
            repo = await InfrahubReadOnlyRepository.init(
                id=self.repository_id,
                name=self.repository_name,
                client=service.client,
                service=service,
            )
        else:
            repo = await InfrahubRepository.init(
                id=self.repository_id,
                name=self.repository_name,
                client=service.client,
                service=service,
            )

        branch = context.branch or repo.default_branch
        commit = repo.get_commit_value(branch_name=branch)

        self._payload = await repo.execute_python_transform.with_options(timeout_seconds=self.transform_timeout)(
            branch_name=branch,
            commit=commit,
            location=f"{self.transform_file}::{self.transform_class}",
            data={"data": data, **context.model_dump()},
            client=service.client,
        )

    @classmethod
    def from_object(cls, obj: CoreCustomWebhook, transform: CoreTransformPython) -> Self:
        return cls(
            name=obj.name.value,
            url=obj.url.value,
            event_type=obj.event_type.value,
            validate_certificates=obj.validate_certificates.value or False,
            repository_id=transform.repository.id,
            repository_name=transform.repository.peer.name.value,
            repository_kind=transform.repository.peer.get_kind(),
            transform_name=transform.name.value,
            transform_class=transform.class_name.value,
            transform_file=transform.file_path.value,
            transform_timeout=transform.timeout.value,
        )



from infrahub.core.constants import AllowOverrideType, BranchSupportType, EventType, InfrahubKind

core_webhook = {
    "name": "Webhook",
    "namespace": "Core",
    "description": "A webhook that connects to an external integration",
    "label": "Webhook",
    "default_filter": "name__value",
    "order_by": ["name__value"],
    "display_labels": ["name__value"],
    "include_in_menu": False,
    "icon": "mdi:webhook",
    "branch": BranchSupportType.AGNOSTIC.value,
    "uniqueness_constraints": [["name__value"]],
    "attributes": [
        {"name": "name", "kind": "Text", "unique": True, "order_weight": 1000},
        {
            "name": "event_type",
            "kind": "Text",
            "enum": ["all"] + EventType.available_types(),
            "default_value": "all",
            "order_weight": 1500,
            "description": "The event type that triggers the webhook",
        },
        {
            "name": "branch_scope",
            "kind": "Dropdown",
            "choices": [
                {
                    "name": "all_branches",
                    "label": "All Branches",
                    "description": "All branches",
                    "color": "#fef08a",
                },
                {
                    "name": "default_branch",
                    "label": "Default Branch",
                    "description": "Only the default branch",
                    "color": "#86efac",
                },
                {
                    "name": "other_branches",
                    "label": "Other Branches",
                    "description": "All branches except the default branch",
                    "color": "#e5e7eb",
                },
            ],
            "default_value": "default_branch",
            "optional": False,
            "order_weight": 2000,
            "allow_override": AllowOverrideType.NONE,
        },
        {"name": "description", "kind": "Text", "optional": True, "order_weight": 2500},
        {"name": "url", "kind": "URL", "order_weight": 3000},
        {
            "name": "validate_certificates",
            "kind": "Boolean",
            "default_value": True,
            "optional": True,
            "order_weight": 5000,
        },
    ],
}

core_standard_webhook = {
    "name": "StandardWebhook",
    "namespace": "Core",
    "description": "A webhook that connects to an external integration",
    "label": "Standard Webhook",
    "default_filter": "name__value",
    "order_by": ["name__value"],
    "display_labels": ["name__value"],
    "include_in_menu": False,
    "icon": "mdi:webhook",
    "branch": BranchSupportType.AGNOSTIC.value,
    "generate_profile": False,
    "inherit_from": [InfrahubKind.WEBHOOK, InfrahubKind.TASKTARGET],
    "attributes": [
        {"name": "shared_key", "kind": "Password", "unique": False, "order_weight": 4000},
    ],
}

core_custom_webhook = {
    "name": "CustomWebhook",
    "namespace": "Core",
    "description": "A webhook that connects to an external integration",
    "label": "Custom Webhook",
    "default_filter": "name__value",
    "order_by": ["name__value"],
    "display_labels": ["name__value"],
    "include_in_menu": False,
    "icon": "mdi:cog-outline",
    "branch": BranchSupportType.AGNOSTIC.value,
    "generate_profile": False,
    "inherit_from": [InfrahubKind.WEBHOOK, InfrahubKind.TASKTARGET],
    "attributes": [],
    "relationships": [
        {
            "name": "transformation",
            "peer": InfrahubKind.TRANSFORMPYTHON,
            "kind": "Attribute",
            "identifier": "webhook___transformation",
            "cardinality": "one",
            "optional": True,
            "order_weight": 7000,
        },
    ],
}



from typing import Any

from infrahub.core.constants import (
    AllowOverrideType,
    BranchSupportType,
    InfrahubKind,
    RepositoryInternalStatus,
    RepositoryOperationalStatus,
    RepositorySyncStatus,
)
from infrahub.core.schema.attribute_schema import AttributeSchema
from infrahub.core.schema.node_schema import NodeSchema

core_repository = NodeSchema(
    name="Repository",
    namespace="Core",
    description="A Git Repository integrated with Infrahub",
    include_in_menu=False,
    icon="mdi:source-repository",
    label="Repository",
    default_filter="name__value",
    order_by=["name__value"],
    display_labels=["name__value"],
    generate_profile=False,
    branch=BranchSupportType.AGNOSTIC,
    inherit_from=[
        InfrahubKind.LINEAGEOWNER,
        InfrahubKind.LINEAGESOURCE,
        InfrahubKind.GENERICREPOSITORY,
        InfrahubKind.TASKTARGET,
    ],
    documentation="/topics/repository",
    attributes=[
        AttributeSchema(
            name="default_branch",
            kind="Text",
            default_value="main",
            order_weight=6000,
        ),
        AttributeSchema(
            name="commit",
            kind="Text",
            optional=True,
            branch=BranchSupportType.LOCAL,
            order_weight=7000,
        ),
    ],
)

core_read_only_repository: dict[str, Any] = {
    "name": "ReadOnlyRepository",
    "namespace": "Core",
    "description": "A Git Repository integrated with Infrahub, Git-side will not be updated",
    "include_in_menu": False,
    "label": "Read-Only Repository",
    "default_filter": "name__value",
    "order_by": ["name__value"],
    "display_labels": ["name__value"],
    "generate_profile": False,
    "branch": BranchSupportType.AGNOSTIC.value,
    "inherit_from": [
        InfrahubKind.LINEAGEOWNER,
        InfrahubKind.LINEAGESOURCE,
        InfrahubKind.GENERICREPOSITORY,
        InfrahubKind.TASKTARGET,
    ],
    "documentation": "/topics/repository",
    "attributes": [
        {
            "name": "ref",
            "kind": "Text",
            "default_value": "main",
            "branch": BranchSupportType.AWARE.value,
            "order_weight": 6000,
        },
        {
            "name": "commit",
            "kind": "Text",
            "optional": True,
            "branch": BranchSupportType.AWARE.value,
            "order_weight": 7000,
        },
    ],
}

core_generic_repository: dict[str, Any] = {
    "name": "GenericRepository",
    "namespace": "Core",
    "label": "Git Repository",
    "description": "A Git Repository integrated with Infrahub",
    "include_in_menu": False,
    "default_filter": "name__value",
    "order_by": ["name__value"],
    "display_labels": ["name__value"],
    "icon": "mdi:source-repository",
    "branch": BranchSupportType.AGNOSTIC.value,
    "uniqueness_constraints": [["name__value"], ["location__value"]],
    "documentation": "/topics/repository",
    "attributes": [
        {
            "name": "name",
            "regex": "^[^/]*$",
            "kind": "Text",
            "unique": True,
            "branch": BranchSupportType.AGNOSTIC.value,
            "order_weight": 1000,
            "allow_override": AllowOverrideType.NONE,
        },
        {
            "name": "description",
            "kind": "Text",
            "optional": True,
            "branch": BranchSupportType.AGNOSTIC.value,
            "order_weight": 2000,
            "allow_override": AllowOverrideType.NONE,
        },
        {
            "name": "location",
            "kind": "Text",
            "unique": True,
            "branch": BranchSupportType.AGNOSTIC.value,
            "order_weight": 3000,
            "allow_override": AllowOverrideType.NONE,
        },
        {
            "name": "internal_status",
            "kind": "Dropdown",
            "choices": [
                {
                    "name": RepositoryInternalStatus.STAGING.value,
                    "label": "Staging",
                    "description": "Repository was recently added to this branch.",
                    "color": "#fef08a",
                },
                {
                    "name": RepositoryInternalStatus.ACTIVE.value,
                    "label": "Active",
                    "description": "Repository is actively being synced for this branch",
                    "color": "#86efac",
                },
                {
                    "name": RepositoryInternalStatus.INACTIVE.value,
                    "label": "Inactive",
                    "description": "Repository is not active on this branch.",
                    "color": "#e5e7eb",
                },
            ],
            "default_value": "inactive",
            "optional": False,
            "branch": BranchSupportType.LOCAL.value,
            "order_weight": 7000,
            "allow_override": AllowOverrideType.NONE,
        },
        {
            "name": "operational_status",
            "kind": "Dropdown",
            "choices": [
                {
                    "name": RepositoryOperationalStatus.UNKNOWN.value,
                    "label": "Unknown",
                    "description": "Status of the repository is unknown and mostlikely because it hasn't been synced yet",
                    "color": "#9ca3af",
                },
                {
                    "name": RepositoryOperationalStatus.ONLINE.value,
                    "label": "Online",
                    "description": "Repository connection is working",
                    "color": "#86efac",
                },
                {
                    "name": RepositoryOperationalStatus.ERROR_CRED.value,
                    "label": "Credential Error",
                    "description": "Repository can't be synced due to some credential error(s)",
                    "color": "#f87171",
                },
                {
                    "name": RepositoryOperationalStatus.ERROR_CONNECTION.value,
                    "label": "Connectivity Error",
                    "description": "Repository can't be synced due to some connectivity error(s)",
                    "color": "#f87171",
                },
                {
                    "name": RepositoryOperationalStatus.ERROR.value,
                    "label": "Error",
                    "description": "Repository can't be synced due to an unknown error",
                    "color": "#ef4444",
                },
            ],
            "optional": False,
            "branch": BranchSupportType.AGNOSTIC.value,
            "default_value": RepositoryOperationalStatus.UNKNOWN.value,
            "order_weight": 5000,
        },
        {
            "name": "sync_status",
            "kind": "Dropdown",
            "choices": [
                {
                    "name": RepositorySyncStatus.UNKNOWN.value,
                    "label": "Unknown",
                    "description": "Status of the repository is unknown and mostlikely because it hasn't been synced yet",
                    "color": "#9ca3af",
                },
                {
                    "name": RepositorySyncStatus.ERROR_IMPORT.value,
                    "label": "Import Error",
                    "description": "Repository import error observed",
                    "color": "#f87171",
                },
                {
                    "name": RepositorySyncStatus.IN_SYNC.value,
                    "label": "In Sync",
                    "description": "The repository is syncing correctly",
                    "color": "#60a5fa",
                },
                {
                    "name": RepositorySyncStatus.SYNCING.value,
                    "label": "Syncing",
                    "description": "A sync job is currently running against the repository.",
                    "color": "#a855f7",
                },
            ],
            "optional": False,
            "branch": BranchSupportType.LOCAL.value,
            "default_value": RepositorySyncStatus.UNKNOWN.value,
            "order_weight": 6000,
        },
    ],
    "relationships": [
        {
            "name": "credential",
            "peer": InfrahubKind.CREDENTIAL,
            "identifier": "gitrepository__credential",
            "kind": "Attribute",
            "optional": True,
            "cardinality": "one",
            "order_weight": 4000,
        },
        {
            "name": "tags",
            "peer": InfrahubKind.TAG,
            "kind": "Attribute",
            "optional": True,
            "cardinality": "many",
            "order_weight": 8000,
        },
        {
            "name": "transformations",
            "peer": InfrahubKind.TRANSFORM,
            "identifier": "repository__transformation",
            "optional": True,
            "cardinality": "many",
            "order_weight": 10000,
        },
        {
            "name": "queries",
            "peer": InfrahubKind.GRAPHQLQUERY,
            "identifier": "graphql_query__repository",
            "optional": True,
            "cardinality": "many",
            "order_weight": 9000,
        },
        {
            "name": "checks",
            "peer": InfrahubKind.CHECKDEFINITION,
            "identifier": "check_definition__repository",
            "optional": True,
            "cardinality": "many",
            "order_weight": 11000,
        },
        {
            "name": "generators",
            "peer": InfrahubKind.GENERATORDEFINITION,
            "identifier": "generator_definition__repository",
            "optional": True,
            "cardinality": "many",
            "order_weight": 12000,
        },
    ],
}



from infrahub.core.constants import (
    BranchSupportType,
    InfrahubKind,
)

core_graphql_query = {
    "name": "GraphQLQuery",
    "namespace": "Core",
    "description": "A pre-defined GraphQL Query",
    "include_in_menu": False,
    "icon": "mdi:graphql",
    "label": "GraphQL Query",
    "default_filter": "name__value",
    "order_by": ["name__value"],
    "display_labels": ["name__value"],
    "generate_profile": False,
    "branch": BranchSupportType.AWARE.value,
    "uniqueness_constraints": [["name__value"]],
    "documentation": "/topics/graphql",
    "attributes": [
        {"name": "name", "kind": "Text", "unique": True},
        {"name": "description", "kind": "Text", "optional": True},
        {"name": "query", "kind": "TextArea"},
        {
            "name": "variables",
            "kind": "JSON",
            "description": "variables in use in the query",
            "optional": True,
            "read_only": True,
        },
        {
            "name": "operations",
            "kind": "List",
            "description": "Operations in use in the query, valid operations: 'query', 'mutation' or 'subscription'",
            "read_only": True,
            "optional": True,
        },
        {
            "name": "models",
            "kind": "List",
            "description": "List of models associated with this query",
            "read_only": True,
            "optional": True,
        },
        {
            "name": "depth",
            "kind": "Number",
            "description": "number of nested levels in the query",
            "read_only": True,
            "optional": True,
        },
        {
            "name": "height",
            "kind": "Number",
            "description": "total number of fields requested in the query",
            "read_only": True,
            "optional": True,
        },
    ],
    "relationships": [
        {
            "name": "repository",
            "peer": InfrahubKind.GENERICREPOSITORY,
            "kind": "Attribute",
            "identifier": "graphql_query__repository",
            "cardinality": "one",
            "optional": True,
        },
        {
            "name": "tags",
            "peer": InfrahubKind.TAG,
            "kind": "Attribute",
            "optional": True,
            "cardinality": "many",
        },
    ],
}



from typing import Union

from prefect import flow

from infrahub.artifacts.models import CheckArtifactCreate
from infrahub.core.constants import InfrahubKind, ValidatorConclusion
from infrahub.core.timestamp import Timestamp
from infrahub.git import InfrahubReadOnlyRepository, InfrahubRepository
from infrahub.services import InfrahubServices
from infrahub.tasks.artifact import define_artifact
from infrahub.workflows.utils import add_tags


@flow(name="git-repository-check-artifact-create", flow_run_name="Check artifact creation")
async def create(model: CheckArtifactCreate, service: InfrahubServices) -> ValidatorConclusion:
    await add_tags(branches=[model.branch_name], nodes=[model.target_id])
    validator = await service.client.get(kind=InfrahubKind.ARTIFACTVALIDATOR, id=model.validator_id, include=["checks"])

    repo: InfrahubReadOnlyRepository | InfrahubRepository
    if InfrahubKind.READONLYREPOSITORY:
        repo = await InfrahubReadOnlyRepository.init(
            id=model.repository_id,
            name=model.repository_name,
            client=service.client,
            service=service,
        )
    else:
        repo = await InfrahubRepository.init(
            id=model.repository_id,
            name=model.repository_name,
            client=service.client,
            service=service,
        )

    artifact, artifact_created = await define_artifact(model=model, service=service)

    severity = "info"
    artifact_result: dict[str, Union[str, bool, None]] = {
        "changed": None,
        "checksum": None,
        "artifact_id": None,
        "storage_id": None,
    }
    check_message = "Failed to render artifact"

    try:
        result = await repo.render_artifact(artifact=artifact, artifact_created=artifact_created, message=model)
        artifact_result["changed"] = result.changed
        artifact_result["checksum"] = result.checksum
        artifact_result["artifact_id"] = result.artifact_id
        artifact_result["storage_id"] = result.storage_id
        check_message = "Artifact rendered successfully"
        conclusion = ValidatorConclusion.SUCCESS

    except Exception as exc:
        artifact.status.value = "Error"
        await artifact.save()
        severity = "critical"
        conclusion = ValidatorConclusion.FAILURE
        check_message += f": {str(exc)}"

    check = None
    check_name = f"{model.artifact_name}: {model.target_name}"
    existing_check = await service.client.filters(
        kind=InfrahubKind.ARTIFACTCHECK, validator__ids=validator.id, name__value=check_name
    )
    if existing_check:
        check = existing_check[0]

    if check:
        check.created_at.value = Timestamp().to_string()
        check.conclusion.value = conclusion.value
        check.severity.value = severity
        check.changed.value = artifact_result["changed"]
        check.checksum.value = artifact_result["checksum"]
        check.artifact_id.value = artifact_result["artifact_id"]
        check.storage_id.value = artifact_result["storage_id"]
        await check.save()
    else:
        check = await service.client.create(
            kind=InfrahubKind.ARTIFACTCHECK,
            data={
                "name": check_name,
                "origin": model.repository_id,
                "kind": "ArtifactDefinition",
                "validator": model.validator_id,
                "created_at": Timestamp().to_string(),
                "message": check_message,
                "conclusion": conclusion.value,
                "severity": severity,
                "changed": artifact_result["changed"],
                "checksum": artifact_result["checksum"],
                "artifact_id": artifact_result["artifact_id"],
                "storage_id": artifact_result["storage_id"],
            },
        )
        await check.save()

    return conclusion



from infrahub.core.constants import (
    BranchSupportType,
    InfrahubKind,
)

core_group = {
    "name": "Group",
    "namespace": "Core",
    "description": "Generic Group Object.",
    "label": "Group",
    "default_filter": "name__value",
    "order_by": ["name__value"],
    "display_labels": ["label__value"],
    "include_in_menu": False,
    "icon": "mdi:group",
    "hierarchical": True,
    "branch": BranchSupportType.AWARE.value,
    "uniqueness_constraints": [["name__value"]],
    "attributes": [
        {"name": "name", "kind": "Text", "unique": True},
        {"name": "label", "kind": "Text", "optional": True},
        {"name": "description", "kind": "Text", "optional": True},
        {
            "name": "group_type",
            "kind": "Text",
            "enum": ["default", "internal"],
            "default_value": "default",
            "optional": False,
        },
    ],
    "relationships": [
        {
            "name": "members",
            "peer": InfrahubKind.NODE,
            "optional": True,
            "identifier": "group_member",
            "cardinality": "many",
            "branch": BranchSupportType.AWARE,
        },
        {
            "name": "subscribers",
            "peer": InfrahubKind.NODE,
            "optional": True,
            "identifier": "group_subscriber",
            "cardinality": "many",
        },
    ],
}

core_standard_group = {
    "name": "StandardGroup",
    "namespace": "Core",
    "description": "Group of nodes of any kind.",
    "include_in_menu": False,
    "icon": "mdi:account-group",
    "label": "Standard Group",
    "default_filter": "name__value",
    "order_by": ["name__value"],
    "display_labels": ["name__value"],
    "branch": BranchSupportType.AWARE.value,
    "inherit_from": [InfrahubKind.GENERICGROUP],
    "generate_profile": False,
}

core_generator_group = {
    "name": "GeneratorGroup",
    "namespace": "Core",
    "description": "Group of nodes that are created by a generator.",
    "include_in_menu": False,
    "icon": "mdi:state-machine",
    "label": "Generator Group",
    "default_filter": "name__value",
    "order_by": ["name__value"],
    "display_labels": ["name__value"],
    "branch": BranchSupportType.LOCAL.value,
    "inherit_from": [InfrahubKind.GENERICGROUP],
    "generate_profile": False,
}

core_graphql_query_group = {
    "name": "GraphQLQueryGroup",
    "namespace": "Core",
    "description": "Group of nodes associated with a given GraphQLQuery.",
    "include_in_menu": False,
    "icon": "mdi:account-group",
    "label": "GraphQL Query Group",
    "default_filter": "name__value",
    "order_by": ["name__value"],
    "display_labels": ["name__value"],
    "branch": BranchSupportType.LOCAL.value,
    "inherit_from": [InfrahubKind.GENERICGROUP],
    "generate_profile": False,
    "attributes": [
        {"name": "parameters", "kind": "JSON", "optional": True},
    ],
    "relationships": [
        {
            "name": "query",
            "peer": InfrahubKind.GRAPHQLQUERY,
            "optional": False,
            "cardinality": "one",
            "kind": "Attribute",
        },
    ],
}



from infrahub.core.constants import (
    ArtifactStatus,
    BranchSupportType,
    ContentType,
    InfrahubKind,
)

core_artifact_target = {
    "name": "ArtifactTarget",
    "include_in_menu": False,
    "namespace": "Core",
    "description": "Extend a node to be associated with artifacts",
    "label": "Artifact Target",
    "relationships": [
        {
            "name": "artifacts",
            "peer": InfrahubKind.ARTIFACT,
            "optional": True,
            "cardinality": "many",
            "kind": "Generic",
            "identifier": "artifact__node",
        },
    ],
}

core_artifact = {
    "name": "Artifact",
    "namespace": "Core",
    "label": "Artifact",
    "include_in_menu": False,
    "icon": "mdi:file-document-outline",
    "default_filter": "name__value",
    "order_by": ["name__value"],
    "display_labels": ["name__value"],
    "branch": BranchSupportType.LOCAL.value,
    "generate_profile": False,
    "inherit_from": [InfrahubKind.TASKTARGET],
    "documentation": "/topics/artifact",
    "attributes": [
        {"name": "name", "kind": "Text"},
        {
            "name": "status",
            "kind": "Text",
            "enum": ArtifactStatus.available_types(),
        },
        {
            "name": "content_type",
            "kind": "Text",
            "enum": ContentType.available_types(),
        },
        {
            "name": "checksum",
            "kind": "Text",
            "optional": True,
        },
        {
            "name": "storage_id",
            "kind": "Text",
            "optional": True,
            "description": "ID of the file in the object store",
        },
        {"name": "parameters", "kind": "JSON", "optional": True},
    ],
    "relationships": [
        {
            "name": "object",
            "peer": InfrahubKind.ARTIFACTTARGET,
            "kind": "Attribute",
            "identifier": "artifact__node",
            "cardinality": "one",
            "optional": False,
        },
        {
            "name": "definition",
            "peer": InfrahubKind.ARTIFACTDEFINITION,
            "kind": "Attribute",
            "identifier": "artifact__artifact_definition",
            "cardinality": "one",
            "optional": False,
        },
    ],
}

core_artifact_definition = {
    "name": "ArtifactDefinition",
    "namespace": "Core",
    "include_in_menu": False,
    "icon": "mdi:file-document-multiple-outline",
    "label": "Artifact Definition",
    "default_filter": "name__value",
    "order_by": ["name__value"],
    "display_labels": ["name__value"],
    "branch": BranchSupportType.AWARE.value,
    "generate_profile": False,
    "uniqueness_constraints": [["name__value"]],
    "inherit_from": [InfrahubKind.TASKTARGET],
    "documentation": "/topics/artifact",
    "attributes": [
        {"name": "name", "kind": "Text", "unique": True},
        {"name": "artifact_name", "kind": "Text"},
        {"name": "description", "kind": "Text", "optional": True},
        {"name": "parameters", "kind": "JSON"},
        {
            "name": "content_type",
            "kind": "Text",
            "enum": ContentType.available_types(),
        },
    ],
    "relationships": [
        {
            "name": "targets",
            "peer": InfrahubKind.GENERICGROUP,
            "kind": "Attribute",
            "identifier": "artifact_definition___group",
            "cardinality": "one",
            "optional": False,
        },
        {
            "name": "transformation",
            "peer": InfrahubKind.TRANSFORM,
            "kind": "Attribute",
            "identifier": "artifact_definition___transformation",
            "cardinality": "one",
            "optional": False,
        },
    ],
}



from typing import Any

core_profile_schema_definition: dict[str, Any] = {
    "name": "Profile",
    "namespace": "Core",
    "include_in_menu": False,
    "icon": "mdi:shape-plus-outline",
    "description": "Base Profile in Infrahub.",
    "label": "Profile",
    "display_labels": ["profile_name__value"],
    "default_filter": "profile_name__value",
    "uniqueness_constraints": [["profile_name__value"]],
    "attributes": [
        {
            "name": "profile_name",
            "kind": "Text",
            "min_length": 3,
            "max_length": 32,
            "optional": False,
            "unique": True,
        },
        {
            "name": "profile_priority",
            "kind": "Number",
            "default_value": 1000,
            "optional": True,
        },
    ],
}



from typing import Any

from infrahub.core.constants import (
    NAMESPACE_REGEX,
)

generic_menu_item: dict[str, Any] = {
    "name": "Menu",
    "namespace": "Core",
    "include_in_menu": False,
    "description": "Element of the Menu",
    "label": "Menu",
    "hierarchical": True,
    "human_friendly_id": ["namespace__value", "name__value"],
    "display_labels": ["label__value"],
    "generate_profile": False,
    "attributes": [
        {"name": "namespace", "kind": "Text", "regex": NAMESPACE_REGEX, "order_weight": 1000},
        {"name": "name", "kind": "Text", "order_weight": 1000},
        {"name": "label", "kind": "Text", "optional": True, "order_weight": 2000},
        {"name": "kind", "kind": "Text", "optional": True, "order_weight": 2500},
        {"name": "path", "kind": "Text", "optional": True, "order_weight": 2500},
        {"name": "description", "kind": "Text", "optional": True, "order_weight": 3000},
        {"name": "icon", "kind": "Text", "optional": True, "order_weight": 4000},
        {"name": "protected", "kind": "Boolean", "default_value": False, "read_only": True, "order_weight": 5000},
        {"name": "order_weight", "kind": "Number", "default_value": 2000, "order_weight": 6000},
        {"name": "required_permissions", "kind": "List", "optional": True, "order_weight": 7000},
        {
            "name": "section",
            "kind": "Text",
            "enum": ["object", "internal"],
            "default_value": "object",
            "order_weight": 8000,
        },
    ],
}

menu_item: dict[str, Any] = {
    "name": "MenuItem",
    "namespace": "Core",
    "include_in_menu": False,
    "description": "Menu Item",
    "label": "Menu Item",
    "inherit_from": ["CoreMenu"],
    "generate_profile": False,
}



from typing import Any

from pydantic import Field, computed_field

from infrahub.core.constants import MutationAction
from infrahub.message_bus import InfrahubMessage

from .constants import EVENT_NAMESPACE
from .models import EventNode, InfrahubEvent


class GroupMutatedEvent(InfrahubEvent):
    """Event generated when a node has been mutated"""

    kind: str = Field(..., description="The type of updated group")
    node_id: str = Field(..., description="The ID of the updated group")
    action: MutationAction = Field(..., description="The action taken on the node")
    members: list[EventNode] = Field(default_factory=list, description="Updated members during this event.")
    ancestors: list[EventNode] = Field(
        default_factory=list, description="A list of groups that are ancestors of this group."
    )

    def get_related(self) -> list[dict[str, str]]:
        related = super().get_related()
        related.append(
            {
                "prefect.resource.id": self.node_id,
                "prefect.resource.role": "infrahub.related.node",
                "infrahub.node.kind": self.kind,
            }
        )
        related.append(
            {
                "prefect.resource.id": self.node_id,
                "prefect.resource.role": "infrahub.group.update",
                "infrahub.node.kind": self.kind,
            }
        )

        for member in self.members:
            related.append(
                {
                    "prefect.resource.id": member.id,
                    "prefect.resource.role": "infrahub.group.member",
                    "infrahub.node.kind": member.kind,
                }
            )
            related.append(
                {
                    "prefect.resource.id": member.id,
                    "prefect.resource.role": "infrahub.related.node",
                    "infrahub.node.kind": member.kind,
                }
            )

        for ancestor in self.ancestors:
            related.append(
                {
                    "prefect.resource.id": ancestor.id,
                    "prefect.resource.role": "infrahub.group.ancestor",
                    "infrahub.node.kind": ancestor.kind,
                }
            )
            related.append(
                {
                    "prefect.resource.id": ancestor.id,
                    "prefect.resource.role": "infrahub.related.node",
                    "infrahub.node.kind": ancestor.kind,
                }
            )
            related.append(
                {
                    "prefect.resource.id": ancestor.id,
                    "prefect.resource.role": "infrahub.group.update",
                    "infrahub.node.kind": ancestor.kind,
                }
            )

        return related

    @computed_field
    def event_name(self) -> str:
        return f"{EVENT_NAMESPACE}.group.{self.action.value}"

    def get_resource(self) -> dict[str, str]:
        return {
            "prefect.resource.id": f"infrahub.node.{self.node_id}",
            "infrahub.node.kind": self.kind,
            "infrahub.node.id": self.node_id,
            "infrahub.node.action": self.action.value,
            "infrahub.node.root_id": self.node_id,
        }

    def get_payload(self) -> dict[str, Any]:
        return {
            "ancestors": [ancestor.model_dump() for ancestor in self.ancestors],
            "members": [member.model_dump() for member in self.members],
        }

    def get_messages(self) -> list[InfrahubMessage]:
        return []


class GroupMemberAddedEvent(GroupMutatedEvent):
    action: MutationAction = MutationAction.CREATED

    @computed_field
    def event_name(self) -> str:
        return f"{EVENT_NAMESPACE}.group.member_added"


class GroupMemberRemovedEvent(GroupMutatedEvent):
    action: MutationAction = MutationAction.DELETED

    @computed_field
    def event_name(self) -> str:
        return f"{EVENT_NAMESPACE}.group.member_removed"



from infrahub.core.constants import (
    DEFAULT_KIND_MAX_LENGTH,
    DEFAULT_KIND_MIN_LENGTH,
    BranchConflictKeep,
    BranchSupportType,
    InfrahubKind,
    RelationshipDeleteBehavior,
    Severity,
    ValidatorConclusion,
    ValidatorState,
)

core_propose_change_validator = {
    "name": "Validator",
    "namespace": "Core",
    "description": "",
    "include_in_menu": False,
    "label": "Validator",
    "order_by": ["started_at__value"],
    "display_labels": ["label__value"],
    "branch": BranchSupportType.AGNOSTIC.value,
    "attributes": [
        {"name": "label", "kind": "Text", "optional": True},
        {
            "name": "state",
            "kind": "Text",
            "enum": ValidatorState.available_types(),
            "default_value": ValidatorState.QUEUED.value,
        },
        {
            "name": "conclusion",
            "kind": "Text",
            "enum": ValidatorConclusion.available_types(),
            "default_value": ValidatorConclusion.UNKNOWN.value,
        },
        {"name": "completed_at", "kind": "DateTime", "optional": True},
        {"name": "started_at", "kind": "DateTime", "optional": True},
    ],
    "relationships": [
        {
            "name": "proposed_change",
            "peer": InfrahubKind.PROPOSEDCHANGE,
            "kind": "Parent",
            "optional": False,
            "cardinality": "one",
            "identifier": "proposed_change__validator",
        },
        {
            "name": "checks",
            "peer": "CoreCheck",
            "kind": "Component",
            "optional": True,
            "cardinality": "many",
            "identifier": "validator__check",
            "on_delete": RelationshipDeleteBehavior.CASCADE,
        },
    ],
}

core_data_validator = {
    "name": "DataValidator",
    "namespace": "Core",
    "description": "A check to validate the data integrity between two branches",
    "include_in_menu": False,
    "label": "Data Validator",
    "display_labels": ["label__value"],
    "inherit_from": [InfrahubKind.VALIDATOR],
    "generate_profile": False,
    "branch": BranchSupportType.AGNOSTIC.value,
}

core_repository_validator = {
    "name": "RepositoryValidator",
    "namespace": "Core",
    "description": "A Validator related to a specific repository",
    "include_in_menu": False,
    "label": "Repository Validator",
    "display_labels": ["label__value"],
    "inherit_from": [InfrahubKind.VALIDATOR],
    "generate_profile": False,
    "branch": BranchSupportType.AGNOSTIC.value,
    "relationships": [
        {
            "name": "repository",
            "peer": InfrahubKind.GENERICREPOSITORY,
            "kind": "Attribute",
            "optional": False,
            "cardinality": "one",
            "branch": BranchSupportType.AGNOSTIC.value,
        },
    ],
}

core_user_validator = {
    "name": "UserValidator",
    "namespace": "Core",
    "description": "A Validator related to a user defined checks in a repository",
    "include_in_menu": False,
    "label": "User Validator",
    "display_labels": ["label__value"],
    "inherit_from": [InfrahubKind.VALIDATOR],
    "generate_profile": False,
    "branch": BranchSupportType.AGNOSTIC.value,
    "relationships": [
        {
            "name": "check_definition",
            "peer": InfrahubKind.CHECKDEFINITION,
            "kind": "Attribute",
            "optional": False,
            "cardinality": "one",
            "branch": BranchSupportType.AGNOSTIC.value,
        },
        {
            "name": "repository",
            "peer": InfrahubKind.GENERICREPOSITORY,
            "kind": "Attribute",
            "optional": False,
            "cardinality": "one",
            "branch": BranchSupportType.AGNOSTIC.value,
        },
    ],
}

core_schema_validator = {
    "name": "SchemaValidator",
    "namespace": "Core",
    "description": "A validator related to the schema",
    "include_in_menu": False,
    "label": "Schema Validator",
    "display_labels": ["label__value"],
    "generate_profile": False,
    "inherit_from": [InfrahubKind.VALIDATOR],
    "branch": BranchSupportType.AGNOSTIC.value,
}

core_artifact_validator = {
    "name": "ArtifactValidator",
    "namespace": "Core",
    "description": "A validator related to the artifacts",
    "include_in_menu": False,
    "label": "Artifact Validator",
    "display_labels": ["label__value"],
    "inherit_from": [InfrahubKind.VALIDATOR],
    "generate_profile": False,
    "branch": BranchSupportType.AGNOSTIC.value,
    "relationships": [
        {
            "name": "definition",
            "peer": InfrahubKind.ARTIFACTDEFINITION,
            "kind": "Attribute",
            "optional": False,
            "cardinality": "one",
            "branch": BranchSupportType.AGNOSTIC.value,
        },
    ],
}

core_generator_validator = {
    "name": "GeneratorValidator",
    "namespace": "Core",
    "description": "A validator related to generators",
    "include_in_menu": False,
    "label": "Generator Validator",
    "display_labels": ["label__value"],
    "inherit_from": [InfrahubKind.VALIDATOR],
    "generate_profile": False,
    "branch": BranchSupportType.AGNOSTIC.value,
    "relationships": [
        {
            "name": "definition",
            "peer": InfrahubKind.GENERATORDEFINITION,
            "kind": "Attribute",
            "optional": False,
            "cardinality": "one",
            "branch": BranchSupportType.AGNOSTIC.value,
        },
    ],
}

core_check = {
    "name": "Check",
    "namespace": "Core",
    "description": "",
    "display_labels": ["label__value"],
    "include_in_menu": False,
    "label": "Check",
    "branch": BranchSupportType.AGNOSTIC.value,
    "attributes": [
        {"name": "name", "kind": "Text", "optional": True},
        {"name": "label", "kind": "Text", "optional": True},
        {"name": "origin", "kind": "Text", "optional": False},
        {
            "name": "kind",
            "kind": "Text",
            "regex": "^[A-Z][a-zA-Z0-9]+$",
            "optional": False,
            "min_length": DEFAULT_KIND_MIN_LENGTH,
            "max_length": DEFAULT_KIND_MAX_LENGTH,
        },
        {"name": "message", "kind": "TextArea", "optional": True},
        {
            "name": "conclusion",
            "kind": "Text",
            "enum": ValidatorConclusion.available_types(),
            "default_value": ValidatorConclusion.UNKNOWN.value,
            "optional": True,
        },
        {
            "name": "severity",
            "kind": "Text",
            "enum": Severity.available_types(),
            "default_value": Severity.INFO.value,
            "optional": True,
        },
        {"name": "created_at", "kind": "DateTime", "optional": True},
    ],
    "relationships": [
        {
            "name": "validator",
            "peer": InfrahubKind.VALIDATOR,
            "identifier": "validator__check",
            "kind": "Parent",
            "optional": False,
            "cardinality": "one",
        },
    ],
}

core_data_check = {
    "name": "DataCheck",
    "namespace": "Core",
    "description": "A check related to some Data",
    "include_in_menu": False,
    "label": "Data Check",
    "display_labels": ["label__value"],
    "inherit_from": ["CoreCheck"],
    "generate_profile": False,
    "branch": BranchSupportType.AGNOSTIC.value,
    "attributes": [
        {"name": "conflicts", "kind": "JSON"},
        {"name": "keep_branch", "enum": BranchConflictKeep.available_types(), "kind": "Text", "optional": True},
        {"name": "enriched_conflict_id", "kind": "Text", "optional": True},
    ],
}

core_standard_check = {
    "name": "StandardCheck",
    "namespace": "Core",
    "description": "A standard check",
    "include_in_menu": False,
    "label": "Standard Check",
    "display_labels": ["label__value"],
    "inherit_from": ["CoreCheck"],
    "generate_profile": False,
    "branch": BranchSupportType.AGNOSTIC.value,
}

core_schema_check = {
    "name": "SchemaCheck",
    "namespace": "Core",
    "description": "A check related to the schema",
    "include_in_menu": False,
    "label": "Schema Check",
    "display_labels": ["label__value"],
    "inherit_from": ["CoreCheck"],
    "generate_profile": False,
    "branch": BranchSupportType.AGNOSTIC.value,
    "attributes": [
        {"name": "conflicts", "kind": "JSON"},
        {"name": "enriched_conflict_id", "kind": "Text", "optional": True},
    ],
}

core_file_check = {
    "name": "FileCheck",
    "namespace": "Core",
    "description": "A check related to a file in a Git Repository",
    "include_in_menu": False,
    "label": "File Check",
    "display_labels": ["label__value"],
    "inherit_from": ["CoreCheck"],
    "generate_profile": False,
    "branch": BranchSupportType.AGNOSTIC.value,
    "attributes": [
        {"name": "files", "kind": "List", "optional": True},
        {"name": "commit", "kind": "Text", "optional": True},
    ],
}

core_artifact_check = {
    "name": "ArtifactCheck",
    "namespace": "Core",
    "description": "A check related to an artifact",
    "include_in_menu": False,
    "label": "Artifact Check",
    "display_labels": ["label__value"],
    "inherit_from": ["CoreCheck"],
    "generate_profile": False,
    "branch": BranchSupportType.AGNOSTIC.value,
    "attributes": [
        {"name": "changed", "kind": "Boolean", "optional": True},
        {"name": "checksum", "kind": "Text", "optional": True},
        {"name": "artifact_id", "kind": "Text", "optional": True},
        {"name": "storage_id", "kind": "Text", "optional": True},
        {"name": "line_number", "kind": "Number", "optional": True},
    ],
}

core_generator_check = {
    "name": "GeneratorCheck",
    "namespace": "Core",
    "description": "A check related to a Generator instance",
    "include_in_menu": False,
    "label": "Generator Check",
    "display_labels": ["label__value"],
    "inherit_from": ["CoreCheck"],
    "generate_profile": False,
    "branch": BranchSupportType.AGNOSTIC.value,
    "attributes": [
        {
            "name": "instance",
            "kind": "Text",
            "optional": False,
        },
    ],
}



from typing import Any

from pydantic import Field, computed_field

from infrahub.core.changelog.models import NodeChangelog
from infrahub.core.constants import MutationAction
from infrahub.message_bus import InfrahubMessage

from .constants import EVENT_NAMESPACE
from .models import InfrahubEvent


class NodeMutatedEvent(InfrahubEvent):
    """Event generated when a node has been mutated"""

    kind: str = Field(..., description="The type of object modified")
    node_id: str = Field(..., description="The ID of the mutated node")
    action: MutationAction = Field(..., description="The action taken on the node")
    data: NodeChangelog = Field(..., description="Data on modified object")
    fields: list[str] = Field(default_factory=list, description="Fields provided in the mutation")

    def get_related(self) -> list[dict[str, str]]:
        related = super().get_related()
        for attribute in self.data.attributes.values():
            related.append(
                {
                    "prefect.resource.id": f"infrahub.node.{self.node_id}",
                    "prefect.resource.role": "infrahub.node.field_update",
                    "infrahub.attribute.name": attribute.name,
                    "infrahub.attribute.value": "NULL" if attribute.value is None else str(attribute.value),
                    "infrahub.attribute.kind": attribute.kind,
                    "infrahub.attribute.value_previous": "NULL"
                    if attribute.value_previous is None
                    else str(attribute.value_previous),
                    # Mypy doesn't understand that .value_update_status is a @computed_attribute
                    "infrahub.attribute.action": attribute.value_update_status.value,  # type: ignore[attr-defined]
                }
            )
        if self.data.parent:
            related.append(
                {
                    "prefect.resource.id": self.data.parent.node_id,
                    "prefect.resource.role": "infrahub.node.parent",
                    "infrahub.parent.kind": self.data.parent.node_kind,
                    "infrahub.parent.id": self.data.parent.node_id,
                }
            )

        related.append(
            {
                "prefect.resource.id": self.node_id,
                "prefect.resource.role": "infrahub.related.node",
                "infrahub.node.kind": self.kind,
            }
        )

        for related_node in self.data.get_related_nodes():
            related.append(
                {
                    "prefect.resource.id": related_node.node_id,
                    "prefect.resource.role": "infrahub.related.node",
                    "infrahub.node.kind": related_node.node_kind,
                }
            )

        return related

    @computed_field
    def event_name(self) -> str:
        return f"{EVENT_NAMESPACE}.node.{self.action.value}"

    def get_resource(self) -> dict[str, str]:
        return {
            "prefect.resource.id": f"infrahub.node.{self.node_id}",
            "infrahub.node.kind": self.kind,
            "infrahub.node.id": self.node_id,
            "infrahub.node.action": self.action.value,
            "infrahub.node.root_id": self.data.root_node_id,
            "infrahub.branch.name": self.meta.context.branch.name,
        }

    def get_payload(self) -> dict[str, Any]:
        return {"data": self.data.model_dump(), "fields": self.fields}

    def get_messages(self) -> list[InfrahubMessage]:
        return [
            # EventNodeMutated(
            #     branch=self.branch,
            #     kind=self.kind,
            #     node_id=self.node_id,
            #     action=self.action.value,
            #     data=self.data,
            #     meta=self.get_message_meta(),
            # )
        ]


class NodeCreatedEvent(NodeMutatedEvent):
    action: MutationAction = MutationAction.CREATED


class NodeUpdatedEvent(NodeMutatedEvent):
    action: MutationAction = MutationAction.UPDATED


class NodeDeletedEvent(NodeMutatedEvent):
    action: MutationAction = MutationAction.DELETED



from infrahub.core.constants import (
    AllowOverrideType,
    BranchSupportType,
    GlobalPermissions,
    InfrahubKind,
    PermissionAction,
    PermissionDecision,
)

core_base_permission = {
    "name": "BasePermission",
    "namespace": "Core",
    "description": "A permission grants right to an account",
    "label": "Base permission",
    "icon": "mdi:user-key",
    "include_in_menu": False,
    "generate_profile": False,
    "attributes": [
        {"name": "description", "kind": "Text", "optional": True},
        {
            "name": "identifier",
            "kind": "Text",
            "read_only": True,
            "optional": True,
            "allow_override": AllowOverrideType.NONE,
        },
    ],
    "relationships": [
        {
            "name": "roles",
            "peer": InfrahubKind.ACCOUNTROLE,
            "optional": True,
            "identifier": "role__permissions",
            "cardinality": "many",
            "kind": "Attribute",
        }
    ],
}

core_object_permission = {
    "name": "ObjectPermission",
    "namespace": "Core",
    "description": "A permission that grants rights to perform actions on objects",
    "label": "Object permission",
    "include_in_menu": False,
    "order_by": ["namespace__value", "name__value", "action__value", "decision__value"],
    "display_labels": ["namespace__value", "name__value", "action__value", "decision__value"],
    "human_friendly_id": ["namespace__value", "name__value", "action__value", "decision__value"],
    "uniqueness_constraints": [["namespace__value", "name__value", "action__value", "decision__value"]],
    "generate_profile": False,
    "inherit_from": [InfrahubKind.BASEPERMISSION],
    "attributes": [
        {"name": "namespace", "kind": "Text", "order_weight": 2000},
        {"name": "name", "kind": "Text", "order_weight": 3000},
        {
            "name": "action",
            "kind": "Text",
            "enum": PermissionAction.available_types(),
            "default_value": PermissionAction.ANY.value,
            "order_weight": 4000,
        },
        {
            "name": "decision",
            "kind": "Number",
            "enum": PermissionDecision.available_types(),
            "default_value": PermissionDecision.ALLOW_ALL.value,
            "order_weight": 5000,
            "description": (
                "Decide to deny or allow the action. If allowed, it can be configured for the default branch, any other branches or all "
                "branches"
            ),
        },
    ],
}

core_global_permission = {
    "name": "GlobalPermission",
    "namespace": "Core",
    "description": "A permission that grants global rights to perform actions in Infrahub",
    "label": "Global permission",
    "include_in_menu": False,
    "order_by": ["action__value", "decision__value"],
    "display_labels": ["action__value", "decision__value"],
    "human_friendly_id": ["action__value", "decision__value"],
    "generate_profile": False,
    "inherit_from": [InfrahubKind.BASEPERMISSION],
    "branch": BranchSupportType.AGNOSTIC.value,
    "attributes": [
        {
            "name": "action",
            "kind": "Dropdown",
            "choices": [{"name": permission.value} for permission in GlobalPermissions],
            "order_weight": 2000,
        },
        {
            "name": "decision",
            "kind": "Number",
            "enum": PermissionDecision.available_types(),
            "default_value": PermissionDecision.ALLOW_ALL.value,
            "order_weight": 3000,
            "description": "Decide to deny or allow the action at a global level",
        },
    ],
}

core_account_role = {
    "name": "AccountRole",
    "namespace": "Core",
    "description": "A role defines a set of permissions to grant to a group of accounts",
    "label": "Account role",
    "icon": "mdi:user-badge",
    "include_in_menu": False,
    "order_by": ["name__value"],
    "display_labels": ["name__value"],
    "human_friendly_id": ["name__value"],
    "generate_profile": False,
    "attributes": [{"name": "name", "kind": "Text", "unique": True}],
    "relationships": [
        {
            "name": "groups",
            "peer": InfrahubKind.ACCOUNTGROUP,
            "optional": True,
            "identifier": "role__accountgroups",
            "cardinality": "many",
            "kind": "Attribute",
        },
        {
            "name": "permissions",
            "peer": InfrahubKind.BASEPERMISSION,
            "optional": True,
            "identifier": "role__permissions",
            "cardinality": "many",
            "kind": "Attribute",
        },
    ],
}

core_account_group = {
    "name": "AccountGroup",
    "namespace": "Core",
    "description": "A group of users to manage common permissions",
    "label": "Account group",
    "icon": "mdi:account-group",
    "include_in_menu": False,
    "order_by": ["name__value"],
    "display_labels": ["name__value"],
    "human_friendly_id": ["name__value"],
    "generate_profile": False,
    "inherit_from": [InfrahubKind.LINEAGEOWNER, InfrahubKind.LINEAGESOURCE, InfrahubKind.GENERICGROUP],
    "branch": BranchSupportType.AGNOSTIC.value,
    "relationships": [
        {
            "name": "roles",
            "peer": InfrahubKind.ACCOUNTROLE,
            "optional": True,
            "identifier": "role__accountgroups",
            "cardinality": "many",
            "kind": "Attribute",
        }
    ],
}
